import itertools
from typing import List, Optional, Sequence, Tuple

import numpy as np
import xarray as xr

from starfish.core.types import Coordinates, Features, Number, OverlapStrategy


class Area:
    """
    Small class that defines rectangular area of physical space by
    its bottom left and top right coordinates.
    """
    def __init__(self, min_x: Number, max_x: Number, min_y: Number, max_y: Number):
        self.min_x = min_x
        self.max_x = max_x
        self.min_y = min_y
        self.max_y = max_y

    def __eq__(self, other) -> bool:
        return (self.min_x == other.min_x
                and self.min_y == other.min_y
                and self.max_x == other.max_x
                and self.max_y == other.max_y)

    @staticmethod
    def _overlap(area1: "Area", area2: "Area") -> bool:
        """Return True if two rectangles overlap"""
        if (area1.max_x < area2.min_x) or (area1.min_x > area2.max_x):
            return False
        if (area1.max_y < area2.min_y) or (area1.min_y > area2.max_y):
            return False
        return True

    @staticmethod
    def find_intersection(area1: "Area", area2: "Area") -> Optional["Area"]:
        """
        Find the overlap area of two rectangles and return as new Area object.
        If the two rectangles do not overlap, return None.
        """
        if Area._overlap(area1, area2):
            return Area(max(area1.min_x, area2.min_x),
                        min(area1.max_x, area2.max_x),
                        max(area1.min_y, area2.min_y),
                        min(area1.max_y, area2.max_y))
        return None


def find_overlaps_of_xarrays(xarrays: Sequence[xr.DataArray]) -> Sequence[Tuple[int, int]]:
    """
    Find all the overlap areas within a list of xarrays.

    Parameters
    ----------
    xarrays : List[xr.DataArray]
        The list of xarrays to find overlaps in.

    Returns
    -------
    List[Tuple[int, int]] :
        A list of tuples containing the indices of two overlapping
        IntensityTables.

    """
    all_overlaps: List[Tuple[int, int]] = list()
    for idx1, idx2 in itertools.combinations(range(len(xarrays)), 2):
        xr1 = xarrays[idx1]
        xr2 = xarrays[idx2]
        area1 = Area(np.min(xr1[Coordinates.X.value]),
                     np.max(xr1[Coordinates.X.value]),
                     np.min(xr1[Coordinates.Y.value]),
                     np.max(xr1[Coordinates.Y.value]))

        area2 = Area(np.min(xr2[Coordinates.X.value]),
                     np.max(xr2[Coordinates.X.value]),
                     np.min(xr2[Coordinates.Y.value]),
                     np.max(xr2[Coordinates.Y.value]))
        if Area._overlap(area1, area2):
            all_overlaps.append((idx1, idx2))
    return all_overlaps


def remove_area_of_xarray(it: xr.DataArray, area: Area) -> xr.DataArray:
    """Return everything in the xarray defined OUTSIDE the input area
    including values on the boundary of the area defined.

    Parameters
    ----------
    it: xr.DataArray
        The xarray to modify
    area: Area
        The area to not include in the modified xarray

    Returns
    -------
    xr.DataArray :
        The xarray without the defined area.
    """
    return it.where((it.xc <= area.min_x)
                    | (it.xc >= area.max_x)
                    | (it.yc <= area.min_y)
                    | (it.yc >= area.max_y),
                    drop=True)


def sel_area_of_xarray(it: xr.DataArray, area: Area) -> xr.DataArray:
    """Return everything in the xarray defined WITHIN the input area
    including values on the boundary of the area defined.

    Parameters
    ----------
    it: xr.DataArray
        The xarray to modify
    area: Area
        The area to include in the modified xarray

    Returns
    -------
     xr.DataArray :
        The xarray within the defined area.
    """
    return it.where((it.xc >= area.min_x)
                    & (it.xc <= area.max_x)
                    & (it.yc >= area.min_y)
                    & (it.yc <= area.max_y), drop=True)


def take_max(it1: xr.DataArray, it2: xr.DataArray):
    """
    Compare two overlapping xarrays and remove spots from whichever has less in the
    overlapping section.

    Parameters
    ----------
    it1 : xr.DataArray
        The first overlapping xarray
    it2 : xr.DataArray
        The second overlapping xarray
    """
    area1 = Area(np.min(it1[Coordinates.X.value]),
                 np.max(it1[Coordinates.X.value]),
                 np.min(it1[Coordinates.Y.value]),
                 np.max(it1[Coordinates.Y.value]))

    area2 = Area(np.min(it2[Coordinates.X.value]),
                 np.max(it2[Coordinates.X.value]),
                 np.min(it2[Coordinates.Y.value]),
                 np.max(it2[Coordinates.Y.value]))
    intersection_rect = Area.find_intersection(area1, area2)
    if not intersection_rect:
        raise ValueError("The given xarrays do not overlap")
    intersect1 = sel_area_of_xarray(it1, intersection_rect)
    intersect2 = sel_area_of_xarray(it2, intersection_rect)
    # compare to see which section has more spots
    if intersect1.sizes[Features.AXIS] >= intersect2.sizes[Features.AXIS]:
        # I1 has more spots remove intesection2 from it2
        it2 = remove_area_of_xarray(it2, intersection_rect)
    else:
        it1 = remove_area_of_xarray(it1, intersection_rect)
    return it1, it2


"""
The mapping between OverlapStrategy type and the method to use for each.
"""
OVERLAP_STRATEGY_MAP = {
    OverlapStrategy.TAKE_MAX: take_max
}



import collections
from typing import Mapping

from starfish.core.types import Axes

_DimensionMetadata = collections.namedtuple("_DimensionMetadata", ['order', 'required'])

AXES_DATA: Mapping[Axes, _DimensionMetadata] = {
    Axes.ROUND: _DimensionMetadata(0, True),
    Axes.CH: _DimensionMetadata(1, True),
    Axes.ZPLANE: _DimensionMetadata(2, False),
}
N_AXES = max(data.order for data in AXES_DATA.values()) + 1



from abc import ABCMeta
from typing import Mapping

import numpy as np
from skimage import img_as_float32
from slicedimage import ImageFormat

from starfish.core.types import Axes
from .all_purpose import LocationAwareFetchedTile

X_COORDS = 0.01, 0.1
Y_COORDS = 0.001, 0.01
Z_COORDS = 0.0001, 0.001


def unique_data(
        fov_id: int, round_label_offset: int, ch_label_offset: int, zplane_label_offset: int,
        num_fovs: int, num_rounds: int, num_chs: int, num_zplanes: int,
        tile_height: int, tile_width: int,
) -> np.ndarray:
    """Return the data for a given tile."""
    result = np.empty((tile_height, tile_width), dtype=np.uint32)
    for row in range(tile_height):
        base_val = tile_width * (
            row + tile_height * (
                zplane_label_offset + num_zplanes * (
                    ch_label_offset + num_chs * (
                        round_label_offset + num_rounds * (
                            fov_id)))))

        result[row:] = np.linspace(base_val, base_val + tile_width, tile_width, False)
    return img_as_float32(result)


class UniqueTiles(LocationAwareFetchedTile, metaclass=ABCMeta):
    """Tiles where the pixel values are unique per round/ch/z."""
    @property
    def format(self) -> ImageFormat:
        return ImageFormat.TIFF

    @property
    def shape(self) -> Mapping[Axes, int]:
        return {Axes.Y: self.tile_height, Axes.X: self.tile_width}

    def tile_data(self) -> np.ndarray:
        """Return the data for a given tile."""
        return unique_data(
            self.fov_id,
            self.rounds.index(self.round_label),
            self.chs.index(self.ch_label),
            self.zplanes.index(self.zplane_label),
            len(self.fovs),
            len(self.rounds),
            len(self.chs),
            len(self.zplanes),
            self.tile_height,
            self.tile_width)



from typing import Tuple, Union

import numpy as np
import pytest
import xarray as xr

from starfish.core.image._filter import element_wise_mult, gaussian_high_pass, mean_high_pass
from starfish.core.imagestack.imagestack import ImageStack
from starfish.core.types import Clip, Number


def random_data_image_stack_factory():
    data = np.random.uniform(0, 1, 100).reshape(1, 1, 1, 10, 10).astype(np.float32)
    return ImageStack.from_numpy(data)


@pytest.mark.parametrize('sigma, is_volume', [
    (1, False),
    ((1, 1), False),
    ((1, 0, 1), True)
])
def test_gaussian_high_pass(sigma: Union[Number, Tuple[Number]], is_volume: bool) -> None:
    """high pass is subtractive, sum of array should be less after running."""
    image_stack = random_data_image_stack_factory()
    sum_before = np.sum(image_stack.xarray)
    ghp = gaussian_high_pass.GaussianHighPass(
        sigma=sigma, is_volume=is_volume, clip_method=Clip.CLIP
    )
    result = ghp.run(image_stack, n_processes=1)
    assert np.sum(result.xarray) < sum_before

@pytest.mark.parametrize('size, is_volume', [
    (1, False),
    ((1, 1), False),
    ((1, 0, 1), True)
])
def test_mean_high_pass(size: Union[Number, Tuple[Number]], is_volume: bool) -> None:
    """high pass is subtractive, sum of array should be less after running."""
    image_stack = random_data_image_stack_factory()
    sum_before = np.sum(image_stack.xarray)
    mhp = mean_high_pass.MeanHighPass(size=size, is_volume=is_volume, clip_method=Clip.CLIP)
    result = mhp.run(image_stack)
    assert np.sum(result.xarray) < sum_before


def test_element_wise_mult() -> None:
    image_stack = random_data_image_stack_factory()
    mult_array = xr.DataArray(
        np.array([[[[[0.5]]]]]),
        dims=('r', 'c', 'z', 'y', 'x')
    )
    ewm = element_wise_mult.ElementWiseMultiply(mult_array)
    multiplied = ewm.run(image_stack, in_place=False)
    assert isinstance(multiplied.xarray, xr.DataArray)
    assert multiplied.xarray.equals(image_stack.xarray * .5)



import os
import subprocess
import sys
from pathlib import Path

import numpy as np

from starfish.core.experiment.experiment import Experiment, FieldOfView


def test_inplace(tmpdir):
    this = Path(__file__)
    inplace_script = this.parent / "inplace_script.py"

    tmpdir_path = Path(tmpdir)

    subprocess.check_call([sys.executable, os.fspath(inplace_script), os.fspath(tmpdir_path)])

    # load up the experiment, and select an image.  Ensure that it has non-zero data.  This is to
    # verify that we are sourcing the data from the tiles that were already on-disk, and not the
    # artificially zero'ed tiles that we feed the experiment builder.
    experiment = Experiment.from_json(os.fspath(tmpdir_path / "experiment.json"))
    primary_image = experiment.fov().get_image(FieldOfView.PRIMARY_IMAGES)
    assert not np.allclose(primary_image.xarray, 0)



import numpy as np

from starfish import data
from starfish.core.image._registration._learn_transform.translation import Translation
from starfish.core.types import Axes


ISS_SHIFTS = [[-23, 6], [-22, 2], [-22, -3], [-15, -4]]


def test_learn_transforms_throws_error():
    exp = data.ISS(use_test_data=True)
    stack = exp.fov().get_image('primary')
    reference_stack = exp.fov().get_image('dots')
    translation = Translation(reference_stack=reference_stack, axes=Axes.ROUND)
    try:
        translation.run(stack)
    except ValueError as e:
        # Assert value error is thrown when the stack is not max projected across all other axes.
        assert e.args[0] == "Only axes: r can have a length > 1, please use the MaxProj filter."


def test_learn_transforms_translation():
    exp = data.ISS(use_test_data=True)
    stack = exp.fov().get_image('primary')
    reference_stack = exp.fov().get_image('dots')
    translation = Translation(reference_stack=reference_stack, axes=Axes.ROUND)
    # Calculate max_proj accrss CH/Z
    stack = stack.max_proj(Axes.CH, Axes.ZPLANE)
    transform_list = translation.run(stack)
    # assert there's a transofrmation object for each round
    assert len(transform_list.transforms) == stack.num_rounds
    for (_, _, transform), shift in zip(transform_list.transforms, ISS_SHIFTS):
        # assert that each TransformationObject has the correct translation shift
        assert np.array_equal(transform.translation, shift)



import collections
import warnings
from copy import deepcopy
from functools import partial
from itertools import product
from json import loads
from pathlib import Path
from typing import (
    Any,
    Callable,
    Iterator,
    List,
    Mapping,
    MutableMapping,
    MutableSequence,
    Optional,
    Sequence,
    Set,
    Tuple,
    Union,
)

import numpy as np
import pandas as pd
import skimage.io
import xarray as xr
from skimage import img_as_float32
from slicedimage import (
    ImageFormat,
    Reader,
    Tile,
    TileSet,
    Writer,
)
from slicedimage.io import resolve_path_or_url
from tqdm import tqdm

from starfish.core.config import StarfishConfig
from starfish.core.errors import DataFormatWarning
from starfish.core.imagestack import indexing_utils
from starfish.core.imagestack.parser import TileCollectionData, TileKey
from starfish.core.imagestack.parser.crop import CropParameters, CroppedTileCollectionData
from starfish.core.imagestack.parser.numpy import NumpyData
from starfish.core.imagestack.parser.tileset import TileSetData
from starfish.core.multiprocessing.pool import Pool
from starfish.core.multiprocessing.shmem import SharedMemory
from starfish.core.types import (
    Axes,
    Clip,
    Coordinates,
    CoordinateValue,
    LOG,
    Number,
    STARFISH_EXTRAS_KEY
)
from starfish.core.util import logging
from starfish.core.util.dtype import preserve_float_range
from ._mp_dataarray import MPDataArray
from .dataorder import AXES_DATA, N_AXES


class ImageStack:
    """
    ImageStacks are the main objects for processing images in starfish. It is a
    5-dimensional Image Tensor that labels each (z, y, x) tile with the round and channel
    and zplane it corresponds to. The class is a wrapper around :py:class:`xarray.DataArray`.
    The names of each imaging r/ch/zplane as well as the physical coordinates of each tile
    are stored as coordinates on the :py:class:`xarray.DataArray`. ImageStacks can only be
    initialized with aligned Tilesets.


    Loads configuration from StarfishConfig.

    Attributes
    ----------
    num_chs : int
        the number of channels stored in the image tensor
    num_rounds : int
        the number of imaging rounds stored in the image tensor
    num_zplanes : int
        the number of z-layers stored in the image tensor
    xarray : :py:class:`xarray.DataArray`
        the 5-d image tensor is stored in this array
    raw_shape : Tuple[int]
        the shape of the image tensor (in integers)
    shape : Dict[str, int]
           the shape of the image tensor by categorical index (channels, imaging rounds, z-layers)
    """

    def __init__(self, data: MPDataArray, tile_data: Optional[TileCollectionData]=None):
        self._data = data
        self._tile_data = tile_data
        self._log: List[dict] = list()

    @classmethod
    def from_tile_collection_data(cls, tile_data: TileCollectionData) -> "ImageStack":
        axes_sizes = {
            Axes.ROUND: len(set(tilekey.round for tilekey in tile_data.keys())),
            Axes.CH: len(set(tilekey.ch for tilekey in tile_data.keys())),
            Axes.ZPLANE: len(set(tilekey.z for tilekey in tile_data.keys())),
        }

        data_shape: MutableSequence[int] = []
        data_dimensions: MutableSequence[str] = []
        data_tick_marks: MutableMapping[str, Sequence[int]] = dict()
        for ix in range(N_AXES):
            size_for_axis: Optional[int] = None
            dim_for_axis: Optional[Axes] = None

            for axis_name, axis_data in AXES_DATA.items():
                if ix == axis_data.order:
                    size_for_axis = axes_sizes[axis_name]
                    dim_for_axis = axis_name
                    break

            if size_for_axis is None or dim_for_axis is None:
                raise ValueError(
                    f"Could not find entry for the {ix}th axis in AXES_DATA")

            data_shape.append(size_for_axis)
            data_dimensions.append(dim_for_axis.value)
            data_tick_marks[dim_for_axis.value] = list(
                sorted(set(tilekey[dim_for_axis] for tilekey in tile_data.keys())))

        data_shape.extend([tile_data.tile_shape[Axes.Y], tile_data.tile_shape[Axes.X]])
        data_dimensions.extend([Axes.Y.value, Axes.X.value])

        # now that we know the tile data type (kind and size), we can allocate the data array.
        data = MPDataArray.from_shape_and_dtype(
            shape=data_shape,
            dtype=np.float32,
            initial_value=0,
            dims=data_dimensions,
            coords=data_tick_marks,
        )

        imagestack = ImageStack(data, tile_data)

        # TODO: (ttung) move more of the initialization code above the constructor call.

        all_selectors = list(imagestack._iter_axes({Axes.ROUND, Axes.CH, Axes.ZPLANE}))
        first_selector = all_selectors[0]
        tile = tile_data.get_tile(r=first_selector[Axes.ROUND],
                                  ch=first_selector[Axes.CH],
                                  z=first_selector[Axes.ZPLANE])

        # Set up coordinates
        imagestack._data[Coordinates.X.value] = xr.DataArray(
            tile.coordinates[Coordinates.X], dims=Axes.X.value)
        imagestack._data[Coordinates.Y.value] = xr.DataArray(
            tile.coordinates[Coordinates.Y], dims=Axes.Y.value)
        # Fill with nan for now, then replace with calculated midpoints
        imagestack._data[Coordinates.Z.value] = xr.DataArray(
            np.full(imagestack.xarray.sizes[Axes.ZPLANE.value], np.nan),
            dims=Axes.ZPLANE.value)

        # Get coords on first tile, then verify all subsequent tiles are aligned
        starting_coords = tile.coordinates

        tile_dtypes = set()
        for selector in tqdm(all_selectors):
            tile = tile_data.get_tile(
                r=selector[Axes.ROUND], ch=selector[Axes.CH], z=selector[Axes.ZPLANE])
            data = tile.numpy_array
            tile_dtypes.add(data.dtype)

            data = img_as_float32(data)
            imagestack.set_slice(selector=selector, data=data)

            if not (
                    np.array_equal(
                        starting_coords[Coordinates.X], tile.coordinates[Coordinates.X])
                    and np.array_equal(
                        starting_coords[Coordinates.Y], tile.coordinates[Coordinates.Y])
            ):
                raise ValueError(f"Tiles must be aligned")
            if Coordinates.Z in tile.coordinates:
                assert len(tile.coordinates[Coordinates.Z]) == 1
                imagestack._data[Coordinates.Z.value].loc[selector[Axes.ZPLANE]] = \
                    tile.coordinates[Coordinates.Z][0]

        tile_dtype_kinds = set(tile_dtype.kind for tile_dtype in tile_dtypes)
        tile_dtype_sizes = set(tile_dtype.itemsize for tile_dtype in tile_dtypes)
        if len(tile_dtype_kinds) != 1:
            raise TypeError("All tiles should have the same kind of dtype")
        if len(tile_dtype_sizes) != 1:
            warnings.warn("Not all tiles have the same precision data", DataFormatWarning)

        # check for existing log info
        if STARFISH_EXTRAS_KEY in tile_data.extras and LOG in tile_data.extras[STARFISH_EXTRAS_KEY]:
            imagestack._log = loads(tile_data.extras[STARFISH_EXTRAS_KEY])[LOG]

        return imagestack

    @staticmethod
    def _validate_data_dtype_and_range(data: Union[np.ndarray, xr.DataArray]) -> None:
        """verify that data is of dtype float32 and in range [0, 1]"""
        if data.dtype != np.float32:
            raise TypeError(
                f"ImageStack data must be of type float32, not {data.dtype}. Please convert data "
                f"using skimage.img_as_float32 prior to calling set_slice."
            )
        if np.min(data) < 0 or np.max(data) > 1:
            raise ValueError(
                f"ImageStack data must be of type float32 and in the range [0, 1]. Please convert "
                f"data using skimage.img_as_float32 prior to calling set_slice."
            )

    def __repr__(self):
        shape = ', '.join(f'{k}: {v}' for k, v in self._data.sizes.items())
        return f"<starfish.ImageStack ({shape})>"

    @classmethod
    def from_tileset(
            cls,
            tileset: TileSet,
            crop_parameters: Optional[CropParameters]=None,
    ) -> "ImageStack":
        """
        Parse a :py:class:`slicedimage.TileSet` into an ImageStack.

        Parameters
        ----------
        tileset : TileSet
            The tileset to parse.
        crop_parameters : Optional[CropParameters]


        Returns
        -------
        ImageStack :
            An ImageStack representing encapsulating the data from the TileSet.
        """
        tile_data: TileCollectionData = TileSetData(tileset)
        if crop_parameters is not None:
            tile_data = CroppedTileCollectionData(tile_data, crop_parameters)
        return ImageStack.from_tile_collection_data(tile_data)

    @classmethod
    def from_url(cls, url: str, baseurl: Optional[str], aligned_group: int = 0):
        """
        Constructs an ImageStack object from a URL and a base URL.

        The following examples will all load from the same location:

        - url: https://www.example.com/images/primary_images.json  baseurl: None
        - url: https://www.example.com/images/primary_images.json  baseurl: I_am_ignored
        - url: primary_images.json  baseurl: https://www.example.com/images
        - url: images/primary_images.json  baseurl: https://www.example.com

        Parameters
        ----------
        url : str
            Either an absolute URL or a relative URL referring to the image to be read.
        baseurl : Optional[str]
            If url is a relative URL, then this must be provided.  If url is an absolute URL, then
            this parameter is ignored.
        aligned_group: int
            Which aligned tile group to load into the Imagestack, only applies if the
            tileset is unaligned. Default 0 (the first group)

        Returns
        -------
        ImageStack :
            An ImageStack representing encapsulating the data from the TileSet.
        """
        config = StarfishConfig()
        tileset = Reader.parse_doc(url, baseurl, backend_config=config.slicedimage)
        coordinate_groups = CropParameters.parse_aligned_groups(tileset)
        crop_params = coordinate_groups[aligned_group]
        return cls.from_tileset(tileset, crop_parameters=crop_params)

    @classmethod
    def from_path_or_url(cls, url_or_path: str, aligned_group: int = 0) -> "ImageStack":
        """
        Constructs an ImageStack object from an absolute URL or a filesystem path.

        The following examples will all load from the same location:

        - url_or_path: file:///Users/starfish-user/images/primary_images.json
        - url_or_path: /Users/starfish-user/images/primary_images.json

        Parameters
        ----------
        url_or_path : str
            Either an absolute URL or a filesystem path to an imagestack.
        aligned_group: int
            Which aligned tile group to load into the Imagestack, only applies if the
            tileset is unaligned. Default 0 (the first group)
        """
        config = StarfishConfig()
        _, relativeurl, baseurl = resolve_path_or_url(url_or_path,
                                                      backend_config=config.slicedimage)
        return cls.from_url(relativeurl, baseurl, aligned_group)

    @classmethod
    def from_numpy(
            cls,
            array: np.ndarray,
            index_labels: Optional[Mapping[Axes, Sequence[int]]]=None,
            coordinates: Optional[Mapping[Coordinates, Sequence[Number]]]=None,
    ) -> "ImageStack":
        """Create an ImageStack from a 5d numpy array with shape (n_round, n_ch, n_z, y, x)

        Parameters
        ----------
        array : np.ndarray
            5-d tensor of shape (n_round, n_ch, n_z, y, x)
        index_labels : Optional[Mapping[Axes, Sequence[int]]]
            Mapping from axes (r, ch, z) to their labels.  If this is not provided, then the axes
            will be labeled from 0..(n-1), where n=the size of the axes.
        coordinates : Optional[Mapping[Coordinates, Sequence[Number]]]
            Map from Coordinates to a sequence of coordinate values.  If this is not provided, then
            the ImageStack gets fake coordinates.

        Returns
        -------
        ImageStack :
            array data stored as an ImageStack

        """
        if len(array.shape) != 5:
            raise ValueError('a 5-d tensor with shape (n_round, n_ch, n_z, y, x) must be provided.')
        try:
            cls._validate_data_dtype_and_range(array)
        except TypeError:
            warnings.warn(f"ImageStack detected as {array.dtype}. Converting to float32...")
            array = img_as_float32(array)

        n_round, n_ch, n_z, height, width = array.shape

        if index_labels is None:
            index_labels = {
                Axes.ROUND: list(range(n_round)),
                Axes.CH: list(range(n_ch)),
                Axes.ZPLANE: list(range(n_z)),
            }
        else:
            assert len(index_labels[Axes.ROUND]) == n_round
            assert len(index_labels[Axes.CH]) == n_ch
            assert len(index_labels[Axes.ZPLANE]) == n_z

        if coordinates is not None:
            assert len(coordinates[Coordinates.X]) == width
            assert len(coordinates[Coordinates.Y]) == height
            assert len(coordinates[Coordinates.Z]) == n_z

        tile_data = NumpyData(array, index_labels, coordinates)
        return ImageStack.from_tile_collection_data(tile_data)

    @property
    def xarray(self) -> xr.DataArray:
        """Retrieves the image data as an :py:class:`xarray.DataArray`"""
        return self._data.data

    def sel(self, indexers: Mapping[Axes, Union[int, tuple]]):
        """Given a dictionary mapping the index name to either a value or a range represented as a
        tuple, return an Imagestack with each dimension indexed accordingly

        Parameters
        ----------
        indexers : Mapping[Axes, Union[int, tuple]]
            A dictionary of dim:index where index is the value or range to index the dimension

        Examples
        --------

        Create an Imagestack :py:func:`~starfish.imagestack.imagestack.ImageStack.synthetic_stack`
            >>> from starfish import ImageStack
            >>> from starfish.core.imagestack.test.factories import synthetic_stack
            >>> from starfish.types import Axes
            >>> stack = synthetic_stack(5, 5, 15, 200, 200)
            >>> stack
            <starfish.ImageStack (r: 5, c: 5, z: 15, y: 200, x: 200)>
            >>> stack.sel({Axes.ROUND: (1, None), Axes.CH: 0, Axes.ZPLANE: 0})
            <starfish.ImageStack (r: 4, c: 1, z: 1, y: 200, x: 200)>
            >>> stack.sel({Axes.ROUND: 0, Axes.CH: 0, Axes.ZPLANE: 1,
            ...Axes.Y: 100, Axes.X: (None, 100)})
            <starfish.ImageStack (r: 1, c: 1, z: 1, y: 1, x: 100)>
            and the imagestack's physical coordinates
            xarray also indexed and recalculated according to the x,y slicing.

        Returns
        -------
        ImageStack :
            a new image stack indexed by given value or range.
        """
        stack = deepcopy(self)
        selector = indexing_utils.convert_to_selector(indexers)
        stack._data._data = indexing_utils.index_keep_dimensions(self.xarray, selector)
        return stack

    def isel(self, indexers: Mapping[Axes, Union[int, tuple]]):
        """Given a dictionary mapping the index name to either a value or a range represented as a
        tuple, return an Imagestack with each dimension indexed by position accordingly

        Parameters
        ----------
        indexers : Dict[Axes, (int/tuple)]
            A dictionary of dim:index where index is the value or range to index the dimension

        Examples
        --------

        Create an Imagestack using the ``synthetic_stack`` method
            >>> from starfish import ImageStack
            >>> from starfish.types import Axes
            >>> stack = ImageStack.synthetic_stack(5, 5, 15, 200, 200)
            >>> stack
            <starfish.ImageStack (r: 5, c: 5, z: 15, y: 200, x: 200)>
            >>> stack.isel({Axes.ROUND: (1, None), Axes.CH: 0, Axes.ZPLANE: 0})
            <starfish.ImageStack (r: 4, c: 1, z: 1, y: 200, x: 200)>
            >>> stack.isel({Axes.ROUND: 0, Axes.CH: 0, Axes.ZPLANE: 1,
            ...Axes.Y: 100, Axes.X: (None, 100)})
            <starfish.ImageStack (r: 1, c: 1, z: 1, y: 1, x: 100)>
            and the imagestack's physical coordinates
            xarray also indexed and recalculated according to the x,y slicing.

        Returns
        -------
        ImageStack :
            a new image stack indexed by given value or range.
        """
        stack = deepcopy(self)
        selector = indexing_utils.convert_to_selector(indexers)
        stack._data._data = indexing_utils.index_keep_dimensions(self.xarray, selector, by_pos=True)
        return stack

    def sel_by_physical_coords(
            self, indexers: Mapping[Coordinates, CoordinateValue]):
        """
        Given a dictionary mapping the coordinate name to either a value or a range represented as a
        tuple, return an Imagestack with each the Coordinate dimension indexed accordingly.

        Parameters
        ----------
        indexers : Mapping[Coordinates, CoordinateValue]:
            A dictionary of coord:index where index is the value or range to index the coordinate
            dimension.

        Returns
        -------
        ImageStack :
            a new image stack indexed by given value or range.
        """
        new_indexers = indexing_utils.convert_coords_to_indices(self.xarray, indexers)
        return self.isel(new_indexers)

    def get_slice(
            self,
            selector: Mapping[Axes, Union[int, slice]]
    ) -> Tuple[np.ndarray, Sequence[Axes]]:
        """
        Given a dictionary mapping the index name to either a value or a slice range, return a
        numpy array representing the slice, and a list of the remaining axes beyond the normal x-y
        tile.

        Examples
        --------

        Slicing with a scalar
            >>> from starfish import ImageStack
            >>> from starfish.core.imagestack.test.factories import synthetic_stack
            >>> from starfish.types import Axes
            >>> stack = synthetic_stack(3, 4, 5, 20, 10)
            >>> stack.shape
            OrderedDict([(<Axes.ROUND: 'r'>, 3),
             (<Axes.CH: 'c'>, 4),
             (<Axes.ZPLANE: 'z'>, 5),
             ('y', 20),
             ('x', 10)])
            >>> stack.axis_labels(Axes.ROUND)
            [0, 1, 2]
            >>> stack.axis_labels(Axes.CH)
            [0, 1, 2, 3]
            >>> stack.axis_labels(Axes.ZPLANE)
            [2, 3, 4, 5, 6]
            >>> data, axes = stack.get_slice({Axes.ZPLANE: 6})
            >>> data.shape
            (3, 4, 20, 10)
            >>> axes
            [<Axes.ROUND: 'r'>, <Axes.CH: 'c'>]

        Slicing with a range
            >>> from starfish import ImageStack
            >>> from starfish.core.imagestack.test.factories import synthetic_stack
            >>> from starfish.types import Axes
            >>> stack = synthetic_stack(3, 4, 5, 20, 10)
            >>> stack.shape
            OrderedDict([(<Axes.ROUND: 'r'>, 3),
             (<Axes.CH: 'c'>, 4),
             (<Axes.ZPLANE: 'z'>, 5),
             ('y', 20),
             ('x', 10)])
            >>> stack.axis_labels(Axes.ROUND)
            [0, 1, 2]
            >>> stack.axis_labels(Axes.CH)
            [0, 1, 2, 3]
            >>> stack.axis_labels(Axes.ZPLANE)
            [2, 3, 4, 5, 6]
            >>> data, axes = stack.get_slice({Axes.ZPLANE: 5, Axes.CH: slice(2, 4)})
            >>> data.shape
            (3, 2, 20, 10)
            >>> axes
            [<Axes.ROUND: 'r'>, <Axes.CH: 'c'>]
        """
        formatted_indexers = indexing_utils.convert_to_selector(selector)
        _, axes = self._build_slice_list(selector)
        result = self._data.sel(formatted_indexers).values

        if result.dtype != np.float32:
            warnings.warn(
                f"Non-float32 dtype: {result.dtype} detected. Data has likely been set using "
                f"private attributes of ImageStack. ImageStack only supports float data in the "
                f"range [0, 1]. Many algorithms will not function properly if provided other "
                f"DataTypes. See: http://scikit-image.org/docs/dev/user_guide/data_types.html")

        return result, axes

    def set_slice(
            self,
            selector: Mapping[Axes, Union[int, slice]],
            data: np.ndarray,
            axes: Optional[Sequence[Axes]]=None):
        """
        Given a dictionary mapping the index name to either a value or a slice range and a source
        numpy array, set the slice of the array of this ImageStack to the values in the source
        numpy array.

        Consumers of this API should not be aware of the internal order of the axes in ImageStack.
        As a result, they should be explicitly providing the order of the axes of the numpy array.
        This method will reorder the data to the internal order of the axes in ImageStack before
        writing it.

        Parameters
        ----------
        selector : Mapping[Axes, Union[int, slice]]
            The slice of the data we are writing with this operation.  Each index should map to a
            value or a range.  If the index is not present, we are writing to the entire range along
            that index.
        data : np.ndarray
            a 2- to 5-D numpy array containing the source data for the operation whose last two axes
            must be in (Y, X) order. If data larger than 2-D is provided, axes must be set to
            specify the order of the additional axes (see below).
        axes : Optional[Sequence[Axes]]
            The order of the axes for the source data, excluding (Y, X). Optional ONLY if data is
            a (Y, X) 2-d tile.

        Examples
        --------
        Setting a slice indicated by scalars.

            >>> import numpy as np
            >>> from starfish import ImageStack
            >>> from starfish.core.imagestack.test.factories import synthetic_stack
            >>> from starfish.types import Axes
            >>> stack = synthetic_stack(3, 4, 5, 20, 10)
            >>> stack.shape
            OrderedDict([(<Axes.ROUND: 'r'>, 3),
             (<Axes.CH: 'c'>, 4),
             (<Axes.ZPLANE: 'z'>, 5),
             ('y', 20),
             ('x', 10)])
            >>> new_data = np.zeros((3, 4, 10, 20), dtype=np.float32)
            >>> stack.set_slice(new_data, axes=[Axes.ROUND, Axes.CH]

        Setting a slice indicated by scalars.  The data presented has a different axis order than
        the previous example.

            >>> import numpy as np
            >>> from starfish import ImageStack
            >>> from starfish.core.imagestack.test.factories import synthetic_stack
            >>> from starfish.types import Axes
            >>> stack = synthetic_stack(3, 4, 5, 20, 10)
            >>> stack.shape
            OrderedDict([(<Axes.ROUND: 'r'>, 3),
             (<Axes.CH: 'c'>, 4),
             (<Axes.ZPLANE: 'z'>, 5),
             ('y', 20),
             ('x', 10)])
            >>> new_data = np.zeros((4, 3, 10, 20), dtype=np.float32)
            >>> stack.set_slice(new_data, axes=[Axes.CH, Axes.ROUND]

        Setting a slice indicated by a range.

            >>> from starfish import ImageStack
            >>> from starfish.core.imagestack.test.factories import synthetic_stack
            >>> from starfish.types import Axes
            >>> stack = synthetic_stack(3, 4, 5, 20, 10)
            >>> stack.shape
            OrderedDict([(<Axes.ROUND: 'r'>, 3),
             (<Axes.CH: 'c'>, 4),
             (<Axes.ZPLANE: 'z'>, 5),
             ('y', 20),
             ('x', 10)])
            >>> new_data = np.zeros((3, 2, 10, 20), dtype=np.float32)
            >>> stack.set_slice({Axes.ZPLANE: 5, Axes.CH: slice(2, 4)}, new_data)
        """

        self._validate_data_dtype_and_range(data)

        slice_list, expected_axes = self._build_slice_list(selector)

        if axes is None:
            axes = list()
        if len(axes) != len(data.shape) - 2:
            raise ValueError(
                "data shape ({}) should be the axes ({}) and (Y,X).".format(data.shape, axes))
        move_src = list()
        move_dst = list()
        for src_idx, axis in enumerate(axes):
            try:
                dst_idx = expected_axes.index(axis)
            except ValueError:
                raise ValueError(
                    "Unexpected axis {}.  Expecting only {}.".format(axis, expected_axes))
            if src_idx != dst_idx:
                move_src.append(src_idx)
                move_dst.append(dst_idx)

        if len(move_src) != 0:
            data = np.moveaxis(data, move_src, move_dst)

        if self._data.loc[slice_list].shape != data.shape:
            raise ValueError("source shape {} mismatches destination shape {}".format(
                data.shape, self._data[slice_list].shape))

        self._data.loc[slice_list] = data

    @staticmethod
    def _build_slice_list(
            selector: Mapping[Axes, Union[int, slice]]
    ) -> Tuple[Tuple[Union[int, slice], ...], Sequence[Axes]]:
        slice_list: MutableSequence[Union[int, slice]] = [
            slice(None, None, None)
            for _ in range(N_AXES)
        ]
        axes = []
        removed_axes = set()
        for name, value in selector.items():
            idx = AXES_DATA[name].order
            if not isinstance(value, slice):
                removed_axes.add(name)
            slice_list[idx] = value

        for dimension_value, dimension_name in sorted([
            (dimension_value.order, dimension_name)
            for dimension_name, dimension_value in AXES_DATA.items()
        ]):
            if dimension_name not in removed_axes:
                axes.append(dimension_name)

        return tuple(slice_list), axes

    def _iter_axes(self, axes: Set[Axes]=None) -> Iterator[Mapping[Axes, int]]:
        """Iterate over provided axes.

        Parameters
        ----------
        axes : Set[Axes]
            The set of Axes to be iterated over (default={Axes.ROUND, Axes.CH}).

        Yields
        ------
        Dict[str, int]
            Mapping of axis name to index

        """
        if axes is None:
            axes = {Axes.ROUND, Axes.CH}
        ordered_axes = list(axes)
        ranges = [self.axis_labels(ind) for ind in ordered_axes]
        for items in product(*ranges):
            a = zip(ordered_axes, items)
            yield {ind: val for (ind, val) in a}

    def apply(
            self,
            func: Callable,
            group_by: Set[Axes]=None,
            in_place=False,
            verbose: bool=False,
            n_processes: Optional[int]=None,
            clip_method: Union[str, Clip]=Clip.CLIP,
            **kwargs
    ) -> "ImageStack":
        """Split the image along a set of axes and apply a function across all the components.  This
        function should yield data of the same dimensionality as the input components.  These
        resulting components are then constituted into an ImageStack and returned.

        Parameters
        ----------
        func : Callable
            Function to apply. must expect a first argument which is a 2d or 3d numpy array and
            return an array of the same shape.
        group_by : Set[Axes]
            Axes to split the data along.
            `ex. splitting a 2D array (axes: X, Y; size:
            3, 4) by X results in 3 arrays of size 4.  (default {Axes.ROUND, Axes.CH,
            Axes.ZPLANE})`
        in_place : bool
            If True, function is executed in place. If n_proc is not 1, the tile or
            volume will be copied once during execution. If false, a new ImageStack object will be
            produced. (Default False)
        verbose : bool
            If True, report on the percentage completed (default = False) during processing
        n_processes : Optional[int]
            The number of processes to use for apply. If None, uses the output of os.cpu_count()
            (default = None).
        kwargs : dict
            Additional arguments to pass to func
        clip_method : Union[str, :py:class:`~starfish.types.Clip`]

            - Clip.CLIP (default): Controls the way that data are scaled to retain skimage dtype
              requirements that float data fall in [0, 1].
            - Clip.CLIP: data above 1 are set to 1, and below 0 are set to 0
            - Clip.SCALE_BY_IMAGE: data above 1 are scaled by the maximum value, with the maximum
              value calculated over the entire ImageStack
            - Clip.SCALE_BY_CHUNK: data above 1 are scaled by the maximum value, with the maximum
              value calculated over each slice, where slice shapes are determined by the group_by
              parameters


        Returns
        -------
        ImageStack :
            If inplace is False, return a new ImageStack, otherwise return a reference to the
            original stack with data modified by application of func

        Raises
        ------
        TypeError :
             If no Clip method given.

        """
        # default grouping is by (x, y) tile
        if group_by is None:
            group_by = {Axes.ROUND, Axes.CH, Axes.ZPLANE}

        if not isinstance(clip_method, (str, Clip)):
            raise TypeError("must pass a Clip method. See starfish.types.Clip for valid options")

        if not in_place:
            # create a copy of the ImageStack, call apply on that stack with in_place=True
            image_stack = deepcopy(self)
            return image_stack.apply(
                func=func,
                group_by=group_by, in_place=True, verbose=verbose, n_processes=n_processes,
                clip_method=clip_method,
                **kwargs
            )

        # wrapper adds a target `data` parameter where the results from func will be stored
        # data are clipped or scaled by chunk using preserve_float_range if clip_method != 2
        bound_func = partial(ImageStack._in_place_apply, func, clip_method=clip_method)

        # execute the processing workflow
        self.transform(
            func=bound_func,
            group_by=group_by,
            verbose=verbose,
            n_processes=n_processes,
            **kwargs)

        # scale based on values of whole image
        if clip_method == Clip.SCALE_BY_IMAGE:
            self._data.data.values = preserve_float_range(self._data.data.values, rescale=True)

        return self

    @staticmethod
    def _in_place_apply(
        apply_func: Callable[..., Union[xr.DataArray, np.ndarray]], data: np.ndarray,
        clip_method: Union[str, Clip], **kwargs
    ) -> None:
        result = apply_func(data, **kwargs)
        if clip_method == Clip.CLIP:
            data[:] = preserve_float_range(result, rescale=False)
        elif clip_method == Clip.SCALE_BY_CHUNK:
            data[:] = preserve_float_range(result, rescale=True)
        else:
            data[:] = result

    def transform(
            self,
            func: Callable,
            group_by: Set[Axes]=None,
            verbose=False,
            n_processes: Optional[int]=None,
            **kwargs
    ) -> List[Any]:
        """Split the image along a set of axes, and apply a function across all the components.

        Parameters
        ----------
        func : Callable
            Function to apply. must expect a first argument which is a numpy array (see group_by)
            but may return any object type.
        group_by : Set[Axes]
            Axes to split the data along.  For instance, splitting a 2D array (axes: X, Y; size:
            3, 4) by X results in 3 arrays of size 4.  (default {Axes.ROUND, Axes.CH,
            Axes.ZPLANE})
        verbose : bool
            If True, report on the percentage completed (default = False) during processing
        n_processes : Optional[int]
            The number of processes to use for apply. If None, uses the output of os.cpu_count()
            (default = None).
        kwargs : dict
            Additional arguments to pass to func being applied

        Returns
        -------
        List[Any] :
            The results of applying func to stored image data
        """
        # default grouping is by (x, y) tile
        if group_by is None:
            group_by = {Axes.ROUND, Axes.CH, Axes.ZPLANE}

        selectors = list(self._iter_axes(group_by))
        slice_lists = [self._build_slice_list(index)[0]
                       for index in selectors]

        selectors_and_slice_lists = zip(selectors, slice_lists)
        if verbose and StarfishConfig().verbose:
            selectors_and_slice_lists = tqdm(selectors_and_slice_lists)

        coordinates = {
            dim: self.xarray.coords[dim]
            for dim in self.xarray.coords.dims
        }

        mp_applyfunc: Callable = partial(
            self._processing_workflow, partial(func, **kwargs), self.xarray.dims, coordinates)

        with Pool(
                processes=n_processes,
                initializer=SharedMemory.initializer,
                initargs=((self._data._backing_mp_array,
                           self._data._data.shape,
                           self._data._data.dtype),)) as pool:
            results = pool.imap(mp_applyfunc, selectors_and_slice_lists)

            # Note: results is [None, ...] if executing an in-place workflow
            # Note: this return must be inside the context manager or the Pool will deadlock
            return list(zip(results, selectors))

    @staticmethod
    def _processing_workflow(
            worker_callable: Callable[[np.ndarray], Any],
            xarray_dims: Sequence[str],
            xarray_coordinates: Mapping[str, np.ndarray],
            selector_and_slice_list: Tuple[Mapping[Axes, int],
                                           Tuple[Union[int, slice], ...]],
    ):
        # build the numpy array from the shared memory object
        backing_mp_array, shape, dtype = SharedMemory.get_payload()
        unshaped_numpy_array = np.frombuffer(backing_mp_array.get_obj(), dtype=dtype)

        # build and then slice the xarray to get the piece needed for this worker
        data_array = xr.DataArray(
            data=unshaped_numpy_array.reshape(shape),
            dims=xarray_dims,
            coords=xarray_coordinates,
        )
        sliced = data_array.sel(selector_and_slice_list[0])

        # pass worker_callable a view into the backing array, which will be overwritten
        return worker_callable(sliced)  # type: ignore

    @property
    def tile_metadata(self) -> pd.DataFrame:
        """return a table containing Tile metadata

        Returns
        -------
        pd.DataFrame :
            dataframe containing per-tile metadata information for each image. Guaranteed to
            include information on channel, imaging round, z plane, and barcode index. Also
            contains any information stored in the extras field for each tile.

        """

        data: collections.defaultdict = collections.defaultdict(list)
        extras_keys: Set[str] = set()
        if self._tile_data is not None:
            tilekeys = self._tile_data.keys()
            extras_keys = set(
                key
                for tilekey in tilekeys
                for key in self._tile_data[tilekey].keys())
        index_keys = set(
            key.value
            for key in AXES_DATA.keys()
        )
        duplicate_keys = index_keys.intersection(extras_keys)
        if len(duplicate_keys) > 0:
            duplicate_keys_str = ", ".join([str(key) for key in duplicate_keys])
            raise ValueError(
                f"keys ({duplicate_keys_str}) was found in both the Tile specification and extras "
                f"field. Tile specification keys may not be duplicated in the extras field.")

        for selector in self._iter_axes({Axes.ROUND, Axes.CH, Axes.ZPLANE}):
            tilekey = TileKey(
                round=selector[Axes.ROUND],
                ch=selector[Axes.CH],
                zplane=selector[Axes.ZPLANE])
            extras = self._tile_data[tilekey] if self._tile_data else {}

            for index, index_value in selector.items():
                data[index.value].append(index_value)

            for k in extras_keys:
                data[k].append(extras.get(k, None))

            if 'barcode_index' not in extras:
                barcode_index = ((((selector[Axes.ZPLANE]
                                    * self.num_rounds) + selector[Axes.ROUND])
                                  * self.num_chs) + selector[Axes.CH])

                data['barcode_index'].append(barcode_index)

        return pd.DataFrame(data)

    @property
    def log(self) -> List[dict]:
        """
        Returns a list of pipeline components that have been applied to this imagestack
        as well as their corresponding runtime parameters.

        For more information about provenance logging see
        `Provenance Logging
        <https://spacetx-starfish.readthedocs.io/en/latest/help_and_reference/api/utils/ilogging.html>`_

        Returns
        -------
        List[dict]
        """
        return self._log

    def update_log(self, class_instance) -> None:
        """
        Adds a new entry to the log list.

        Parameters
        ----------
        class_instance: The instance of a class being applied to the imagestack
        """
        entry = {"method": class_instance.__class__.__name__,
                 "arguments": class_instance.__dict__,
                 "os": logging.get_os_info(),
                 "dependencies": logging.get_core_dependency_info(),
                 "release tag": logging.get_release_tag(),
                 "starfish version": logging.get_dependency_version('starfish')
                 }
        self._log.append(entry)

    @property
    def raw_shape(self) -> Tuple[int, int, int, int, int]:
        """
        Returns the shape of the 5-d image tensor stored as self.image

        Returns
        -------
        Tuple[int, int, int, int, int] :
            The size of the image tensor
        """
        return self._data.shape

    @property
    def shape(self) -> collections.OrderedDict:
        """
        Returns the shape of the space that this image inhabits.  It does not include the
        dimensions of the image itself.  For instance, if this is an X-Y image in a C-H-Y-X space,
        then the shape would include the axes C and H.

        Returns
        -------
        An ordered mapping between index names to the size of the index.
        """
        # TODO: (ttung) Note that the return type should be ..OrderedDict[Any, str], but python3.6
        # has a bug where this # breaks horribly.  Can't find a bug id to link to, but see
        # https://stackoverflow.com/questions/41207128/how-do-i-specify-ordereddict-k-v-types-for-\
        # mypy-type-annotation
        result: collections.OrderedDict[Any, str] = collections.OrderedDict()
        for name, data in AXES_DATA.items():
            result[name] = self._data.shape[data.order]
        result['y'] = self._data.shape[-2]
        result['x'] = self._data.shape[-1]

        return result

    @property
    def num_rounds(self):
        """Return the number of rounds in the ImageStack"""
        return self.xarray.sizes[Axes.ROUND]

    @property
    def num_chs(self):
        """Return the number of channels in the ImageStack"""
        return self.xarray.sizes[Axes.CH]

    @property
    def num_zplanes(self):
        """Return the number of z_planes in the ImageStack"""
        return self.xarray.sizes[Axes.ZPLANE]

    def axis_labels(self, axis: Axes) -> Sequence[int]:
        """Given an axis, return the sorted unique values for that axis in this ImageStack.  For
        instance, ``imagestack.axis_labels(Axes.ROUND)`` returns all the round ids in this
        imagestack."""

        return [int(val) for val in self.xarray.coords[axis.value].values]

    @property
    def tile_shape(self):
        """Return the shape of each tile in the ImageStack. All
        Tiles have the same shape."""
        return self.xarray.sizes[Axes.Y], self.xarray.sizes[Axes.X]

    def to_multipage_tiff(self, filepath: str) -> None:
        """save the ImageStack as a FIJI-compatible multi-page TIFF file

        Parameters
        ----------
        filepath : str
            filepath for a tiff FILE. "TIFF" suffix will be added if the provided path does not
            end with .TIFF

        """
        if not filepath.upper().endswith(".TIFF"):
            filepath += ".TIFF"

        # RZCYX is the order expected by FIJI
        data = self.xarray.transpose(
            Axes.ROUND.value,
            Axes.ZPLANE.value,
            Axes.CH.value,
            Axes.Y.value,
            Axes.X.value)

        # Any float32 image with low dynamic range will provoke a warning that the image is
        # low contrast because the data must be converted to uint16 for compatibility with FIJI.
        with warnings.catch_warnings():
            warnings.simplefilter("ignore", UserWarning)
            skimage.io.imsave(filepath, data.values, imagej=True)

    def export(self,
               filepath: str,
               tile_opener=None,
               tile_format: ImageFormat=ImageFormat.NUMPY) -> None:
        """write the image tensor to disk in spaceTx format

        Parameters
        ----------
        filepath : str
            Path + prefix for the images and primary_images.json written by this function
        tile_opener : TODO ttung: doc me.
        tile_format : ImageFormat
            Format in which each 2D plane should be written.

        """
        # Add log data to extras
        tileset_extras = self._tile_data.extras if self._tile_data else {}
        tileset_extras[STARFISH_EXTRAS_KEY] = logging.LogEncoder().encode({LOG: self.log})
        tileset = TileSet(
            dimensions={
                Axes.ROUND,
                Axes.CH,
                Axes.ZPLANE,
                Axes.Y,
                Axes.X,
            },
            shape={
                Axes.ROUND: self.num_rounds,
                Axes.CH: self.num_chs,
                Axes.ZPLANE: self.num_zplanes,
            },
            default_tile_shape={Axes.Y: self.tile_shape[0], Axes.X: self.tile_shape[1]},
            extras=tileset_extras,
        )
        for selector in self._iter_axes({Axes.ROUND, Axes.CH, Axes.ZPLANE}):
            tilekey = TileKey(
                round=selector[Axes.ROUND],
                ch=selector[Axes.CH],
                zplane=selector[Axes.ZPLANE])
            extras: dict = self._tile_data[tilekey] if self._tile_data else {}

            coordinates: MutableMapping[Coordinates, Union[Tuple[Number, Number], Number]] = dict()
            x_coordinates = (float(self.xarray[Coordinates.X.value][0]),
                             float(self.xarray[Coordinates.X.value][-1]))
            y_coordinates = (float(self.xarray[Coordinates.Y.value][0]),
                             float(self.xarray[Coordinates.Y.value][-1]))
            coordinates[Coordinates.X] = x_coordinates
            coordinates[Coordinates.Y] = y_coordinates
            if Coordinates.Z in self.xarray.coords:
                # set the z coord to the calculated value from the associated z plane
                z_coordinates = float(self.xarray[Coordinates.Z.value][selector[Axes.ZPLANE]])
                coordinates[Coordinates.Z] = z_coordinates

            tile = Tile(
                coordinates=coordinates,
                indices=selector,
                extras=extras,
            )
            tile.numpy_array, _ = self.get_slice(selector)
            tileset.add_tile(tile)

        if tile_opener is None:
            def tile_opener(tileset_path: Path, tile, ext):
                base = tileset_path.parent / tileset_path.stem
                if Axes.ZPLANE in tile.indices:
                    zval = tile.indices[Axes.ZPLANE]
                    zstr = "-Z{}".format(zval)
                else:
                    zstr = ""
                return open(
                    "{}-H{}-C{}{}.{}".format(
                        str(base),
                        tile.indices[Axes.ROUND],
                        tile.indices[Axes.CH],
                        zstr,
                        ext,
                    ),
                    "wb")

        if not filepath.endswith('.json'):
            filepath += '.json'
        Writer.write_to_path(
            tileset,
            filepath,
            pretty=True,
            tile_opener=tile_opener,
            tile_format=tile_format)

    def max_proj(self, *dims: Axes) -> "ImageStack":
        """return a max projection over one or more axis of the image tensor

        Parameters
        ----------
        dims : Axes
            one or more axes to project over

        Returns
        -------
        np.ndarray :
            max projection

        """
        max_projection = self._data.max([dim.value for dim in dims])
        max_projection = max_projection.expand_dims(tuple(dim.value for dim in dims))
        max_projection = max_projection.transpose(*self.xarray.dims)
        physical_coords: MutableMapping[Coordinates, Sequence[Number]] = {}
        for axis, coord in (
                (Axes.X, Coordinates.X),
                (Axes.Y, Coordinates.Y),
                (Axes.ZPLANE, Coordinates.Z)):
            if axis in dims:
                # this axis was projected out of existence.
                assert coord.value not in max_projection.coords
                physical_coords[coord] = [np.average(self._data.coords[coord.value])]
            else:
                physical_coords[coord] = max_projection.coords[coord.value]
        max_proj_stack = ImageStack.from_numpy(max_projection.values, coordinates=physical_coords)
        return max_proj_stack

    def _squeezed_numpy(self, *dims: Axes):
        """return this ImageStack's data as a squeezed numpy array"""
        return self.xarray.squeeze(tuple(dim.value for dim in dims)).values



"""
Tests for IntensityTable.from_image_stack method
"""

import numpy as np

from starfish import ImageStack
from starfish.core.imagestack.test import test_labeled_indices
from starfish.core.test.factories import (
    codebook_intensities_image_for_single_synthetic_spot,
    synthetic_spot_pass_through_stack,
)
from starfish.core.types import Axes
from ..intensity_table import IntensityTable


# TODO ambrosejcarr: improve the tests here.
def test_imagestack_to_intensity_table():
    codebook, intensity_table, image = codebook_intensities_image_for_single_synthetic_spot()
    pixel_intensities = IntensityTable.from_image_stack(image)
    pixel_intensities = codebook.decode_metric(
        pixel_intensities, max_distance=0, min_intensity=1000, norm_order=2)
    assert isinstance(pixel_intensities, IntensityTable)


def test_imagestack_to_intensity_table_no_noise():
    codebook, intensity_table, image = synthetic_spot_pass_through_stack()
    pixel_intensities = IntensityTable.from_image_stack(image)
    pixel_intensities = codebook.decode_metric(
        pixel_intensities, max_distance=0, min_intensity=1000, norm_order=2)
    assert isinstance(pixel_intensities, IntensityTable)


# TODO ambrosejcarr: crop is not tested because it should be moved out of this function
def test_intensity_table_can_be_constructed_from_an_imagestack():
    """
    ImageStack has enough information to create an IntensityTable without additional SpotAttributes.
    Each feature is a pixel, and therefore the SpotAttributes can be extracted from the relative
    locations.
    """
    r, c, z, y, x = 1, 5, 2, 2, 5
    data = np.zeros(100, dtype=np.float32).reshape(r, c, z, y, x)
    image_stack = ImageStack.from_numpy(data)
    intensities = IntensityTable.from_image_stack(image_stack)

    # there should be 100 features
    assert np.product(intensities.shape) == 100

    # the max features should be equal to the array extent (2, 2, 5) minus one, since indices
    # are being compared and python is zero based
    # import pdb; pdb.set_trace()
    assert np.max(intensities[Axes.ZPLANE.value].values) == z - 1
    assert np.max(intensities[Axes.Y.value].values) == y - 1
    assert np.max(intensities[Axes.X.value].values) == x - 1

    # the number of channels and rounds should match the ImageStack
    assert intensities.sizes[Axes.CH.value] == c
    assert intensities.sizes[Axes.ROUND.value] == r


def test_from_imagestack_labeled_indices():
    # use the ImageStack with labeled indices from the test.
    imagestack = test_labeled_indices.setup_imagestack()
    intensity_table = IntensityTable.from_image_stack(imagestack)
    assert np.array_equal(
        intensity_table[Axes.CH.value], np.array(test_labeled_indices.CH_LABELS))
    assert np.array_equal(
        intensity_table[Axes.ROUND.value], np.array(test_labeled_indices.ROUND_LABELS))



import hashlib
import os
import sys
from pathlib import Path
from typing import Mapping, Union

import numpy as np
from skimage.io import imsave
from slicedimage import ImageFormat

from starfish.core.experiment.builder import FetchedTile, TileFetcher, write_experiment_json
from starfish.core.experiment.builder.inplace import (
    enable_inplace_mode, inplace_tile_opener, InplaceFetchedTile
)
from starfish.core.experiment.experiment import Experiment, FieldOfView
from starfish.core.types import Axes, Coordinates, CoordinateValue


SHAPE = {Axes.Y: 500, Axes.X: 1390}


def tile_fn(input_dir: Path, prefix: str, fov: int, r: int, ch: int, zplane: int) -> Path:
    filename = '{}-Z{}-H{}-C{}.tiff'.format(prefix, zplane, r, ch)
    return input_dir / f"fov_{fov:03}" / filename


class ZeroesInplaceTile(InplaceFetchedTile):
    """These tiles contain all zeroes.  This is irrelevant to the actual experiment construction
    because we are using in-place mode.  That means we build references to the files already on-disk
    and any data returned is merely metadata (tile shape, tile coordinates, and tile checksum."""
    def __init__(self, file_path: Path):
        self.file_path = file_path

    @property
    def shape(self) -> Mapping[Axes, int]:
        return SHAPE

    @property
    def coordinates(self) -> Mapping[Union[str, Coordinates], CoordinateValue]:
        return {
            Coordinates.X: (0.0, 0.0001),
            Coordinates.Y: (0.0, 0.0001),
            Coordinates.Z: (0.0, 0.0001),
        }

    @property
    def sha256(self) -> str:
        hasher = hashlib.sha256()
        with open(str(self.file_path), "rb") as fh:
            hasher.update(fh.read())
        return hasher.hexdigest()

    @property
    def filepath(self) -> Path:
        return self.file_path


class ZeroesInplaceFetcher(TileFetcher):
    def __init__(self, input_dir: Path, prefix: str):
        self.input_dir = input_dir
        self.prefix = prefix

    def get_tile(
            self, fov_id: int, round_label: int, ch_label: int, zplane_label: int) -> FetchedTile:
        filename = '{}-Z{}-H{}-C{}.tiff'.format(self.prefix, zplane_label, round_label, ch_label)
        file_path = self.input_dir / f"fov_{fov_id:03}" / filename
        return ZeroesInplaceTile(file_path)


def fov_path_generator(parent_toc_path: Path, toc_name: str) -> Path:
    return parent_toc_path.parent / toc_name / "{}.json".format(parent_toc_path.stem)


def format_data(
        image_dir: Path,
        primary_image_dimensions: Mapping[Union[Axes, str], int],
        aux_name_to_dimensions: Mapping[str, Mapping[Union[Axes, str], int]],
        num_fovs):
    def add_codebook(experiment_json_doc):
        experiment_json_doc['codebook'] = "codebook.json"
        return experiment_json_doc

    enable_inplace_mode()

    write_experiment_json(
        path=os.fspath(image_dir),
        fov_count=num_fovs,
        tile_format=ImageFormat.TIFF,
        primary_image_dimensions=primary_image_dimensions,
        aux_name_to_dimensions=aux_name_to_dimensions,
        primary_tile_fetcher=ZeroesInplaceFetcher(image_dir, FieldOfView.PRIMARY_IMAGES),
        aux_tile_fetcher={
            aux_img_name: ZeroesInplaceFetcher(image_dir, aux_img_name)
            for aux_img_name in aux_name_to_dimensions.keys()
        },
        postprocess_func=add_codebook,
        default_shape=SHAPE,
        fov_path_generator=fov_path_generator,
        tile_opener=inplace_tile_opener,
    )


def write_image(
        base_path: Path,
        prefix: str,
        num_fovs: int,
        image_dimensions: Mapping[Union[Axes, str], int],
):
    """Writes the constituent tiles of an image to disk.  The tiles are made up with random noise.
    """
    for fov_num in range(num_fovs):
        for round_label in range(image_dimensions[Axes.ROUND]):
            for ch_label in range(image_dimensions[Axes.CH]):
                for zplane_label in range(image_dimensions[Axes.ZPLANE]):
                    path = tile_fn(base_path, prefix, fov_num, round_label, ch_label, zplane_label)
                    path.parent.mkdir(parents=True, exist_ok=True)

                    data = np.random.random(size=(SHAPE[Axes.Y], SHAPE[Axes.X])).astype(np.float32)

                    imsave(os.fspath(path), data, plugin="tifffile")


def write_inplace(path: str, num_fovs: int = 2):
    primary_image_dimensions: Mapping[Union[Axes, str], int] = {
        Axes.ROUND: 4,
        Axes.CH: 4,
        Axes.ZPLANE: 1,
    }

    aux_name_to_dimensions: Mapping[str, Mapping[Union[Axes, str], int]] = {
        'nuclei': {
            Axes.ROUND: 1,
            Axes.CH: 1,
            Axes.ZPLANE: 1,
        },
    }

    tmpdir = Path(path)

    # write out the image files
    write_image(tmpdir, FieldOfView.PRIMARY_IMAGES, num_fovs, primary_image_dimensions)

    for aux_img_name in aux_name_to_dimensions.keys():
        write_image(tmpdir, aux_img_name, num_fovs, aux_name_to_dimensions[aux_img_name])

    # format the experiment.
    format_data(tmpdir, primary_image_dimensions, aux_name_to_dimensions, num_fovs)

    Experiment.from_json(os.fspath(tmpdir / "experiment.json"))


if __name__ == "__main__":
    write_inplace(sys.argv[1])



import warnings
from typing import Optional, Tuple, Union

import numpy as np
import xarray as xr
from trackpy import locate

from starfish.core.imagestack.imagestack import ImageStack
from starfish.core.intensity_table.intensity_table import IntensityTable
from starfish.core.types import Axes, SpotAttributes
from starfish.core.util import click
from ._base import DetectSpotsAlgorithmBase
from .detect import detect_spots


class TrackpyLocalMaxPeakFinder(DetectSpotsAlgorithmBase):
    """
    Find spots using a local max peak finding algorithm

    This is a wrapper for :code:`trackpy.locate`, which implements a version of the
    `Crocker-Grier <crocker_grier>`_ algorithm.

    .. _crocker_grier: https://physics.nyu.edu/grierlab/methods3c/

    Parameters
    ----------

    spot_diameter : odd integer or tuple of odd integers.
        This may be a single number or a tuple giving the feature’s extent in each dimension,
        useful when the dimensions do not have equal resolution (e.g. confocal microscopy).
        The tuple order is the same as the image shape, conventionally (z, y, x) or (y, x).
        The number(s) must be odd integers. When in doubt, round up.
    min_mass : float, optional
        The minimum integrated brightness. This is a crucial parameter for eliminating spurious
        features. Recommended minimum values are 100 for integer images and 1 for float images.
        Defaults to 0 (no filtering).
    max_size : float
        maximum radius-of-gyration of brightness, default None
    separation : float or tuple
        Minimum separtion between features. Default is diameter + 1. May be a tuple, see
        diameter for details.
    percentile : float
        Features must have a peak brighter than pixels in this percentile. This helps eliminate
        spurious peaks.
    noise_size : float or tuple
        Width of Gaussian blurring kernel, in pixels Default is 1. May be a tuple, see diameter
        for details.
    smoothing_size : float or tuple
        The size of the sides of the square kernel used in boxcar (rolling average) smoothing,
        in pixels Default is diameter. May be a tuple, making the kernel rectangular.
    threshold : float
        Clip bandpass result below this value. Thresholding is done on the already
        background-subtracted image. By default, 1 for integer images and 1/255 for float
        images.
    measurement_type : str ['max', 'mean']
        name of the function used to calculate the intensity for each identified spot area
    preprocess : boolean
        Set to False to turn off bandpass preprocessing.
    max_iterations : integer
        Max number of loops to refine the center of mass, default 10
    is_volume : bool
        if True, run the algorithm on 3d volumes of the provided stack
    verbose : bool
        If True, report the percentage completed (default = False) during processing

    Notes
    -----
    See also trackpy locate: http://soft-matter.github.io/trackpy/dev/generated/trackpy.locate.html

    """

    def __init__(
            self, spot_diameter, min_mass, max_size, separation, percentile=0,
            noise_size: Tuple[int, int, int] = (1, 1, 1), smoothing_size=None, threshold=None,
            preprocess: bool = False, max_iterations: int = 10, measurement_type: str = 'max',
            is_volume: bool = False, verbose=False) -> None:

        self.diameter = spot_diameter
        self.minmass = min_mass
        self.maxsize = max_size
        self.separation = separation
        self.noise_size = noise_size
        self.smoothing_size = smoothing_size
        self.percentile = percentile
        self.threshold = threshold
        self.measurement_function = self._get_measurement_function(measurement_type)
        self.preprocess = preprocess
        self.max_iterations = max_iterations
        self.is_volume = is_volume
        self.verbose = verbose

    def image_to_spots(self, data_image: Union[np.ndarray, xr.DataArray]) -> SpotAttributes:
        """

        Parameters
        ----------
        data_image : np.ndarray
            three-dimensional image containing spots to be detected

        Returns
        -------
        SpotAttributes :
            spot attributes table for all detected spots

        """
        data_image = np.asarray(data_image)
        with warnings.catch_warnings():
            warnings.simplefilter('ignore', FutureWarning)  # trackpy numpy indexing warning
            warnings.simplefilter('ignore', UserWarning)  # yielded if black images
            attributes = locate(
                data_image,
                diameter=self.diameter,
                minmass=self.minmass,
                maxsize=self.maxsize,
                separation=self.separation,
                noise_size=self.noise_size,
                smoothing_size=self.smoothing_size,
                threshold=self.threshold,
                percentile=self.percentile,
                preprocess=self.preprocess,
                max_iterations=self.max_iterations,
            )

        # when zero spots are detected, 'ep' is missing from the trackpy locate results.
        if attributes.shape[0] == 0:
            attributes['ep'] = []

        # TODO ambrosejcarr: data should always be at least pseudo-3d, this may not be necessary
        # TODO ambrosejcarr: this is where max vs. sum vs. mean would be parametrized.
        # here, total_intensity = sum, intensity = max
        new_colnames = [
            'y', 'x', 'total_intensity', 'radius', 'eccentricity', 'intensity', 'raw_mass', 'ep'
        ]
        if len(data_image.shape) == 3:
            attributes.columns = ['z'] + new_colnames
        else:
            attributes.columns = new_colnames

        attributes['spot_id'] = np.arange(attributes.shape[0])
        return SpotAttributes(attributes)

    def run(
            self,
            primary_image: ImageStack,
            blobs_image: Optional[ImageStack] = None,
            blobs_axes: Optional[Tuple[Axes, ...]] = None,
            n_processes: Optional[int] = None,
            *args,
    ) -> IntensityTable:
        """
        Find spots.

        Parameters
        ----------
        primary_image : ImageStack
            ImageStack where we find the spots in.
        blobs_image : Optional[ImageStack]
            If provided, spots will be found in the blobs image, and intensities will be measured
            across rounds and channels. Otherwise, spots are measured independently for each channel
            and round.
        blobs_axes : Optional[Tuple[Axes, ...]]
            If blobs_image is provided, blobs_axes must be provided as well.  blobs_axes represents
            the axes across which the blobs image is max projected before spot detection is done.
        n_processes : Optional[int] = None,
            Number of processes to devote to spot finding.
        """
        intensity_table = detect_spots(
            data_stack=primary_image,
            spot_finding_method=self.image_to_spots,
            reference_image=blobs_image,
            reference_image_max_projection_axes=blobs_axes,
            measurement_function=self.measurement_function,
            n_processes=n_processes,
            radius_is_gyration=True)

        return intensity_table

    @staticmethod
    @click.command("TrackpyLocalMaxPeakFinder")
    @click.option("--spot-diameter", type=str, help='expected spot size')
    @click.option(
        "--min-mass", default=4, type=int, help="minimum integrated spot intensity")
    @click.option(
        "--max-size", default=6, type=int, help="maximum radius of gyration of brightness")
    @click.option(
        "--separation", default=5, type=float, help="minimum distance between spots")
    @click.option(
        "--noise-size", default=None, type=int,
        help="width of gaussian blurring kernel, in pixels")
    @click.option(
        "--smoothing-size", default=None, type=int,
        help="odd integer. Size of boxcar (moving average) filter in pixels. Default is the "
             "Diameter")
    @click.option(
        "--preprocess", is_flag=True,
        help="if passed, gaussian and boxcar filtering are applied")
    @click.option(
        "--max-iterations", default=10, type=int,
        help="Max number of loops to refine the center of mass. Default is 10")
    @click.option(
        "--show", default=False, is_flag=True, help="display results visually")
    @click.option(
        "--percentile", default=None, type=float,
        help="clip bandpass below this value. Thresholding is done on already background-"
             "subtracted images. Default 1 for integer images and 1/255 for float")
    @click.option(
        "--is-volume", is_flag=True,
        help="indicates that the image stack should be filtered in 3d")
    @click.pass_context
    def _cli(ctx, spot_diameter, min_max, max_size, separation, noise_size, smoothing_size,
             preprocess, max_iterations, show, percentile, is_volume):

        instance = TrackpyLocalMaxPeakFinder(spot_diameter, min_max, max_size,
                                             separation, noise_size, smoothing_size,
                                             preprocess, max_iterations, show,
                                             percentile, is_volume)
        ctx.obj["component"]._cli_run(ctx, instance)



"""Simple utility function we commonly use across ImageStack tests."""
from typing import Mapping, Optional, Sequence, Tuple, Union

import numpy as np

from starfish.core.types import Axes, Coordinates, Number
from ..imagestack import ImageStack


def verify_stack_data(
        stack: ImageStack,
        selectors: Mapping[Axes, Union[int, slice]],
        expected_data: np.ndarray,
) -> Tuple[np.ndarray, Sequence[Axes]]:
    """Given an imagestack and a set of selectors, verify that the data referred to by the selectors
    matches the expected data.
    """
    tile_data, axes = stack.get_slice(selectors)
    assert np.array_equal(tile_data, expected_data)

    return tile_data, axes


def verify_physical_coordinates(
        stack: ImageStack,
        expected_x_coordinates: Tuple[Number, Number],
        expected_y_coordinates: Tuple[Number, Number],
        expected_z_coordinates: Union[Number, Tuple[Number, Number]],
        zplane: Optional[int] = None,
) -> None:
    """Given an imagestack and a set of coordinate min/max values verify that the physical
    coordinates on the stack match the expected range of values for each coord dimension.
    """
    assert np.all(np.isclose(stack.xarray[Coordinates.X.value],
                             np.linspace(expected_x_coordinates[0],
                                         expected_x_coordinates[1],
                                         stack.xarray.sizes[Axes.X.value])))
    assert np.all(np.isclose(stack.xarray[Coordinates.Y.value],
                             np.linspace(expected_y_coordinates[0],
                                         expected_y_coordinates[1],
                                         stack.xarray.sizes[Axes.Y.value])))
    # If zplane provided, test expected_z_coordinates on specific plane.
    # Else just test expected_z_coordinates on entire array
    if zplane is not None:
        assert np.isclose(
            stack.xarray.sel({Axes.ZPLANE.value: zplane})[Coordinates.Z.value],
            expected_z_coordinates)
    else:
        assert np.all(np.isclose(stack.xarray[Coordinates.Z.value], expected_z_coordinates))


def _calculate_physical_pixel_size(coord_min: Number, coord_max: Number, num_pixels: int) -> Number:
    """Calculate the size of a pixel in physical space"""
    return (coord_max - coord_min) / num_pixels


def _pixel_offset_to_physical_coordinate(
        physical_pixel_size: Number,
        pixel_offset: Optional[int],
        coordinates_at_pixel_offset_0: Number,
) -> Number:
    """Calculate the physical pixel value at the given index"""
    if pixel_offset:
        # Check for negative index
        assert pixel_offset >= 0
        return (physical_pixel_size * pixel_offset) + coordinates_at_pixel_offset_0
    return coordinates_at_pixel_offset_0


def recalculate_physical_coordinate_range(
        coord_min: float,
        coord_max: float,
        dimension_size: int,
        indexer: Union[int, slice],
) -> Tuple[Number, Number]:
    """Given the dimension size and pixel coordinate indexes calculate the corresponding
    coordinates in physical space

    Parameters
    ----------
    coord_min: float
        the minimum physical coordinate value

    coord_max: float
        the maximum physical coordinate value

    dimension_size: int
        The size (number of pixels) to use to calculate physical_pixel_size

    indexer: (int/slice)
        The pixel index or range to calculate.

    Returns
    -------
    The new min and max physical coordinate values of the given dimension
    """
    physical_pixel_size = _calculate_physical_pixel_size(coord_min, coord_max, dimension_size - 1)
    min_pixel_index = indexer if isinstance(indexer, int) else indexer.start
    if isinstance(indexer, int):
        max_pixel_index = indexer
    elif isinstance(indexer.stop, int):
        if indexer.stop >= 0:
            max_pixel_index = indexer.stop - 1
        else:
            max_pixel_index = indexer.stop + dimension_size
    else:
        max_pixel_index = dimension_size - 1
    new_min = _pixel_offset_to_physical_coordinate(physical_pixel_size, min_pixel_index, coord_min)
    new_max = _pixel_offset_to_physical_coordinate(physical_pixel_size, max_pixel_index, coord_min)
    return new_min, new_max



from starfish.core.codebook.codebook import Codebook
from starfish.core.intensity_table.intensity_table import IntensityTable
from starfish.core.util import click
from ._base import DecodeAlgorithmBase


class PerRoundMaxChannel(DecodeAlgorithmBase):
    """
    Decode spots by selecting the max-valued channel in each sequencing round.

    Note that this assumes that the codebook contains only one "on" channel per sequencing round,
    a common pattern in experiments that assign one fluorophore to each DNA nucleotide and
    read DNA sequentially. It is also a characteristic of single-molecule FISH and RNAscope
    codebooks.

    Parameters
    ----------
    codebook : Codebook
        Contains codes to decode IntensityTable

    """

    def __init__(self, codebook: Codebook):
        self.codebook = codebook

    def run(self, intensities: IntensityTable, *args) -> IntensityTable:
        """Decode spots by selecting the max-valued channel in each sequencing round

        Parameters
        ----------
        intensities : IntensityTable
            IntensityTable to be decoded

        Returns
        -------
        IntensityTable :
            IntensityTable decoded and appended with Features.TARGET and Features.QUALITY values.

        """
        return self.codebook.decode_per_round_max(intensities)

    @staticmethod
    @click.command("PerRoundMaxChannel")
    @click.pass_context
    def _cli(ctx):
        codebook = ctx.obj["codebook"]
        ctx.obj["component"]._cli_run(ctx, PerRoundMaxChannel(codebook))



import warnings

import numpy as np
import pytest

from starfish import ImageStack
from .fakefilter import (
    AdditiveFilterAlgorithm,
    FilterAlgorithmWithMissingConstructorTyping,
    FilterAlgorithmWithMissingRunTyping,
    SimpleFilterAlgorithm,
)
from ..errors import (
    ConstructorError,
    ConstructorExtraParameterWarning,
    ExecutionError,
    RunInsufficientParametersError,
    TypeInferenceError,
)
from ..filesystem import FileProvider
from ..runnable import Runnable


BASE_EXPECTED = np.array([
    [0.227543, 0.223117, 0.217014, 0.221241, 0.212863, 0.211963, 0.210575,
     0.198611, 0.194827, 0.181964],
    [0.216617, 0.214710, 0.212467, 0.218158, 0.211429, 0.210361, 0.205737,
     0.190814, 0.182010, 0.165667],
    [0.206744, 0.204685, 0.208774, 0.212909, 0.215274, 0.206180, 0.196674,
     0.179080, 0.169207, 0.157549],
    [0.190845, 0.197131, 0.188540, 0.195361, 0.196765, 0.200153, 0.183627,
     0.167590, 0.159930, 0.150805],
    [0.181231, 0.187457, 0.182910, 0.179416, 0.175357, 0.172137, 0.165072,
     0.156344, 0.153735, 0.150378],
    [0.169924, 0.184604, 0.182422, 0.174441, 0.159823, 0.157229, 0.157259,
     0.151690, 0.147265, 0.139940],
    [0.164874, 0.169467, 0.178012, 0.173129, 0.161425, 0.155978, 0.152712,
     0.150286, 0.145159, 0.140658],
    [0.164508, 0.165042, 0.171420, 0.174990, 0.162951, 0.152422, 0.149325,
     0.151675, 0.141588, 0.139010],
    [0.162448, 0.156451, 0.158419, 0.162722, 0.160388, 0.152865, 0.142885,
     0.142123, 0.140093, 0.135836],
    [0.150072, 0.147295, 0.145495, 0.153216, 0.156085, 0.149981, 0.145571,
     0.141878, 0.138857, 0.136965]],
    dtype=np.float32)
URL = "https://d2nhj9g34unfro.cloudfront.net/20181005/ISS-TEST/fov_001/hybridization.json"


def test_str():
    """Verify that we can get a sane string for a runnable."""
    filter_runnable = Runnable(
        SimpleFilterAlgorithm,
        FileProvider(URL),
        multiplicand=.5,
    )
    assert str(filter_runnable) == ("compute(Filter.SimpleFilterAlgorithm,"
                                    + f" FileProvider(\"{URL}\"), multiplicand=0.5)")


def test_constructor_error():
    """Verify that we get a properly typed error when the constructor does not execute correctly.
    In this case, we do not provide enough parameters to `SimpleFilterAlgorithm`."""
    with pytest.raises(ConstructorError):
        Runnable(
            SimpleFilterAlgorithm,
            FileProvider(URL),
        )


def test_execution_error():
    """Verify that we get a properly typed error when the constructor does not execute correctly."""
    filter_runnable = Runnable(
        SimpleFilterAlgorithm,
        FileProvider(URL),
        multiplicand="abacadabra",
    )
    with pytest.raises(ExecutionError):
        filter_runnable.run({})


def test_constructor_type_inference_error():
    """Verify that we get a properly typed error when we cannot properly infer the type for one of
    the constructor's parameters."""
    with pytest.raises(TypeInferenceError):
        Runnable(
            FilterAlgorithmWithMissingConstructorTyping,
            FileProvider(URL),
            additive=FileProvider(URL),
        )


def test_run_type_inference_error():
    """Verify that we get a properly typed error when we cannot properly infer the type for one of
    the run method's parameters.  In this case, `FilterAlgorithmWithMissingRunTyping` does not
    provide a type hint for one of its constructor arguments."""
    with pytest.raises(TypeInferenceError):
        Runnable(
            FilterAlgorithmWithMissingRunTyping,
            FileProvider(URL),
            multiplicand=FileProvider(URL),
        )


def test_extra_constructor_parameter_fileprovider():
    """Verify that we raise a warning when we provide extra parameters that are fileproviders to an
    algorithm's constructor."""
    with warnings.catch_warnings(record=True) as w:
        filter_runnable = Runnable(
            SimpleFilterAlgorithm,
            FileProvider(URL),
            multiplicand=.5,
            additive=FileProvider(URL),
        )
        assert len(w) == 1
        assert issubclass(w[-1].category, ConstructorExtraParameterWarning)

    result = filter_runnable.run({})
    assert isinstance(result, ImageStack)

    # pick a random part of the filtered image and assert on it
    assert result.xarray.dtype == np.float32

    assert np.allclose(
        BASE_EXPECTED * .5,
        result.xarray[2, 2, 0, 40:50, 40:50]
    )


def test_extra_constructor_parameter_non_fileprovider():
    """Verify that we raise a warning when we provide extra parameters that are not fileproviders
    to an algorithm's constructor."""
    with warnings.catch_warnings(record=True) as w:
        filter_runnable = Runnable(
            SimpleFilterAlgorithm,
            FileProvider(URL),
            multiplicand=.5,
            additive=.5,
        )
        assert len(w) == 1
        assert issubclass(w[-1].category, ConstructorExtraParameterWarning)

    result = filter_runnable.run({})
    assert isinstance(result, ImageStack)

    # pick a random part of the filtered image and assert on it
    assert result.xarray.dtype == np.float32

    assert np.allclose(
        BASE_EXPECTED * .5,
        result.xarray[2, 2, 0, 40:50, 40:50]
    )


def test_run_insufficient_parameters():
    """Verify that we can run a single runnable and get its result.
    """
    with pytest.raises(RunInsufficientParametersError):
        Runnable(
            SimpleFilterAlgorithm,
            multiplicand=.5,
        )


def test_run():
    """Verify that we can run a single runnable and get its result.
    """
    filter_runnable = Runnable(
        SimpleFilterAlgorithm,
        FileProvider(URL),
        multiplicand=.5,
    )
    result = filter_runnable.run({})
    assert isinstance(result, ImageStack)

    # pick a random part of the filtered image and assert on it
    assert result.xarray.dtype == np.float32

    assert np.allclose(
        BASE_EXPECTED * .5,
        result.xarray[2, 2, 0, 40:50, 40:50]
    )


def test_chained_run():
    """Verify that we can run a runnable that depends on another runnable.
    """
    dependency_runnable = Runnable(
        SimpleFilterAlgorithm,
        FileProvider(URL),
        multiplicand=.5,
    )
    result = dependency_runnable.run({})
    assert isinstance(result, ImageStack)

    filter_runnable = Runnable(
        SimpleFilterAlgorithm,
        dependency_runnable,
        multiplicand=2.0,
    )
    result = filter_runnable.run({dependency_runnable: result})
    assert isinstance(result, ImageStack)

    # pick a random part of the filtered image and assert on it
    assert result.xarray.dtype == np.float32

    assert np.allclose(
        BASE_EXPECTED,
        result.xarray[2, 2, 0, 40:50, 40:50]
    )


def test_chained_run_result_not_present():
    """Verify that we can run a runnable that depends on another runnable, but the results are not
    present.
    """
    dependency_runnable = Runnable(
        SimpleFilterAlgorithm,
        FileProvider(URL),
        multiplicand=.5,
    )
    result = dependency_runnable.run({})
    assert isinstance(result, ImageStack)

    filter_runnable = Runnable(
        SimpleFilterAlgorithm,
        dependency_runnable,
        multiplicand=2.0,
    )
    with pytest.raises(KeyError):
        filter_runnable.run({})


def test_load_data_for_constructor():
    """Verify that we can properly load up data from a FileProvider that is required for the
    constructor."""
    filter_runnable = Runnable(
        AdditiveFilterAlgorithm,
        FileProvider(URL),
        additive=FileProvider(URL),
    )
    result = filter_runnable.run({})
    assert isinstance(result, ImageStack)

    # pick a random part of the filtered image and assert on it
    assert result.xarray.dtype == np.float32

    assert np.allclose(
        BASE_EXPECTED * 2,
        result.xarray[2, 2, 0, 40:50, 40:50]
    )



"""
Notes
-----
This test and docs/source/usage/iss/iss_cli.sh test the same code paths and should be updated
together
"""
import os
import unittest

import numpy as np
import pandas as pd
import pytest

from starfish.types import Features
from ._base_cli_test import CLITest


EXPERIMENT_JSON_URL = "https://d2nhj9g34unfro.cloudfront.net/20181005/ISS-TEST/experiment.json"


@pytest.mark.slow
class TestWithIssData(CLITest, unittest.TestCase):

    @property
    def spots_file(self):
        return "decoded-spots.nc"

    @property
    def subdirs(self):
        return (
            "max_projected",
            "transforms",
            "registered",
            "filtered",
            "results",
        )

    @property
    def stages(self):
        return (
            [
                "starfish", "validate", "experiment", EXPERIMENT_JSON_URL,
            ],
            [
                "starfish", "Filter",
                "--input",
                f"@{EXPERIMENT_JSON_URL}[fov_001][primary]",
                "--output", lambda tempdir, *args, **kwargs: os.path.join(
                    tempdir, "max_projected", "primary_images.json"),
                "MaxProject",
                "--dims", "c",
                "--dims", "z"

            ],
            [
                "starfish", "LearnTransform",
                "--input", lambda tempdir, *args, **kwargs: os.path.join(
                    tempdir, "max_projected", "primary_images.json"),
                "--output", lambda tempdir, *args, **kwargs: os.path.join(
                    tempdir, "transforms", "transforms.json"),
                "Translation",
                "--reference-stack",
                f"@{EXPERIMENT_JSON_URL}[fov_001][dots]",
                "--upsampling", "1000",
                "--axes", "r"
            ],
            [
                "starfish", "ApplyTransform",
                "--input",
                f"@{EXPERIMENT_JSON_URL}[fov_001][primary]",
                "--output", lambda tempdir, *args, **kwargs: os.path.join(
                    tempdir, "registered", "primary_images.json"),
                "--transformation-list", lambda tempdir, *args, **kwargs: os.path.join(
                    tempdir, "transforms", "transforms.json"),
                "Warp",
            ],
            [
                "starfish", "Filter",
                "--input", lambda tempdir, *args, **kwargs: os.path.join(
                    tempdir, "registered", "primary_images.json"),
                "--output", lambda tempdir, *args, **kwargs: os.path.join(
                    tempdir, "filtered", "primary_images.json"),
                "WhiteTophat",
                "--masking-radius", "15",
            ],
            [
                "starfish", "Filter",
                "--input",
                f"@{EXPERIMENT_JSON_URL}[fov_001][nuclei]",
                "--output", lambda tempdir, *args, **kwargs: os.path.join(
                    tempdir, "filtered", "nuclei.json"),
                "WhiteTophat",
                "--masking-radius", "15",
            ],
            [
                "starfish", "Filter",
                "--input",
                f"@{EXPERIMENT_JSON_URL}[fov_001][dots]",
                "--output", lambda tempdir, *args, **kwargs: os.path.join(
                    tempdir, "filtered", "dots.json"),
                "WhiteTophat",
                "--masking-radius", "15",
            ],
            [
                "starfish", "DetectSpots",
                "--input", lambda tempdir, *args, **kwargs: os.path.join(
                    tempdir, "filtered", "primary_images.json"),
                "--output", lambda tempdir, *args, **kwargs: os.path.join(
                    tempdir, "results", "spots.nc"),
                "--blobs-stack", lambda tempdir, *args, **kwargs: os.path.join(
                    tempdir, "filtered", "dots.json"),
                "--blobs-axis", "r", "--blobs-axis", "c",
                "BlobDetector",
                "--min-sigma", "4",
                "--max-sigma", "6",
                "--num-sigma", "20",
                "--threshold", "0.01",
            ],
            [
                "starfish", "Segment",
                "--primary-images", lambda tempdir, *args, **kwargs: os.path.join(
                    tempdir, "filtered", "primary_images.json"),
                "--nuclei", lambda tempdir, *args, **kwargs: os.path.join(
                    tempdir, "filtered", "nuclei.json"),
                "-o", lambda tempdir, *args, **kwargs: os.path.join(
                    tempdir, "results", "masks.tgz"),
                "Watershed",
                "--nuclei-threshold", ".16",
                "--input-threshold", ".22",
                "--min-distance", "57",
            ],
            [
                "starfish", "AssignTargets",
                "--label-image",
                lambda tempdir, *args, **kwargs: os.path.join(
                    tempdir, "results", "masks.tgz"),
                "--intensities", lambda tempdir, *args, **kwargs: os.path.join(
                    tempdir, "results", "spots.nc"),
                "--output", lambda tempdir, *args, **kwargs: os.path.join(
                    tempdir, "results", "targeted-spots.nc"),
                "Label",
            ],
            [
                "starfish", "Decode",
                "-i", lambda tempdir, *args, **kwargs: os.path.join(
                    tempdir, "results", "targeted-spots.nc"),
                "--codebook",
                f"@{EXPERIMENT_JSON_URL}",
                "-o", lambda tempdir, *args, **kwargs: os.path.join(
                    tempdir, "results", "decoded-spots.nc"),
                "PerRoundMaxChannel",
            ],

            # Validate results/{spots,targeted-spots,decoded-spots}.nc
            [
                "starfish", "validate", "xarray",
                lambda tempdir, *args, **kwargs: os.path.join(
                    tempdir, "results", "spots.nc")
            ],
            [
                "starfish", "validate", "xarray",
                lambda tempdir, *args, **kwargs: os.path.join(
                    tempdir, "results", "targeted-spots.nc")
            ],
            [
                "starfish", "validate", "xarray",
                lambda tempdir, *args, **kwargs: os.path.join(
                    tempdir, "results", "decoded-spots.nc")
            ],
        )

    def verify_results(self, intensities):
        # TODO make this test stronger
        genes, counts = np.unique(
            intensities.coords[Features.TARGET], return_counts=True)
        gene_counts = pd.Series(counts, genes)
        # TODO THERE"S NO HUMAN/MOUSE KEYS?
        assert gene_counts['ACTB']



import warnings
from typing import Iterable, Optional, Union

from starfish.core.imagestack.imagestack import ImageStack
from starfish.core.types import Axes
from starfish.core.util import click
from ._base import FilterAlgorithmBase


class MaxProject(FilterAlgorithmBase):
    """
    Creates a maximum projection over one or more axis of the image tensor

    Parameters
    ----------
    dims : Axes
        one or more Axes to project over

    See Also
    --------
    starfish.types.Axes

    """

    def __init__(self, dims: Iterable[Union[Axes, str]]) -> None:

        warnings.warn(
            "Filter.MaxProject is being deprecated in favor of Filter.Reduce(func='max')",
            DeprecationWarning,
        )
        self.dims = dims

    _DEFAULT_TESTING_PARAMETERS = {"dims": 'r'}

    def run(
            self,
            stack: ImageStack,
            in_place: bool = False,
            verbose: bool = False,
            n_processes: Optional[int] = None,
            *args,
    ) -> ImageStack:
        """Perform filtering of an image stack

        Parameters
        ----------
        stack : ImageStack
            Stack to be filtered.
        in_place : bool
            if True, process ImageStack in-place, otherwise return a new stack
        verbose : bool
            if True, report on filtering progress (default = False)
        n_processes : Optional[int]
            Number of parallel processes to devote to calculating the filter

        Returns
        -------
        ImageStack :
            If in-place is False, return the results of filter as a new stack. Otherwise return the
            original stack.

        """
        return stack.max_proj(*tuple(Axes(dim) for dim in self.dims))

    @staticmethod
    @click.command("MaxProject")
    @click.option(
        "--dims",
        type=click.Choice(
            [Axes.ROUND.value, Axes.CH.value, Axes.ZPLANE.value, Axes.X.value, Axes.Y.value]
        ),
        multiple=True,
        help="The dimensions the Imagestack should max project over."
             "For multiple dimensions add multiple --dims. Ex."
             "--dims r --dims c")
    @click.pass_context
    def _cli(ctx, dims):
        ctx.obj["component"]._cli_run(ctx, MaxProject(dims))



import os

import numpy as np
import pytest
import xarray as xr

from starfish.core.segmentation_mask import (
    SegmentationMaskCollection,
    validate_segmentation_mask,
)
from starfish.core.types import Axes, Coordinates


def test_validate_segmentation_mask():
    good = xr.DataArray([[True, False, False],
                         [False, True, True]],
                        dims=('y', 'x'),
                        coords=dict(x=[0, 1, 2],
                                    y=[0, 1],
                                    xc=('x', [0.5, 1.5, 2.5]),
                                    yc=('y', [0.5, 1.5])))
    validate_segmentation_mask(good)

    good = xr.DataArray([[[True], [False], [False]],
                         [[False], [True], [True]]],
                        dims=('z', 'y', 'x'),
                        coords=dict(z=[0, 1],
                                    y=[1, 2, 3],
                                    x=[42],
                                    zc=('z', [0.5, 1.5]),
                                    yc=('y', [1.5, 2.5, 3.5]),
                                    xc=('x', [42.5])))
    validate_segmentation_mask(good)

    bad = xr.DataArray([[1, 2, 3],
                        [4, 5, 6]],
                       dims=('y', 'x'),
                       coords=dict(x=[0, 1, 2],
                                   y=[0, 1],
                                   xc=('x', [0.5, 1.5, 2.5]),
                                   yc=('y', [0.5, 1.5])))
    with pytest.raises(TypeError):
        validate_segmentation_mask(bad)

    bad = xr.DataArray([True],
                       dims=('x'),
                       coords=dict(x=[0],
                                   xc=('x', [0.5])))
    with pytest.raises(TypeError):
        validate_segmentation_mask(bad)

    bad = xr.DataArray([[True]],
                       dims=('z', 'y'),
                       coords=dict(z=[0],
                                   y=[0],
                                   zc=('z', [0.5]),
                                   yc=('y', [0.5])))
    with pytest.raises(TypeError):
        validate_segmentation_mask(bad)

    bad = xr.DataArray([[True]],
                       dims=('x', 'y'))
    with pytest.raises(TypeError):
        validate_segmentation_mask(bad)


def test_from_label_image():
    label_image = np.zeros((5, 5), dtype=np.int32)
    label_image[0] = 1
    label_image[3:5, 3:5] = 2
    label_image[-1, -1] = 0

    physical_ticks = {Coordinates.Y: [1.2, 2.4, 3.6, 4.8, 6.0],
                      Coordinates.X: [7.2, 8.4, 9.6, 10.8, 12]}

    masks = SegmentationMaskCollection.from_label_image(label_image,
                                                        physical_ticks)

    assert len(masks) == 2

    region_1, region_2 = masks

    assert region_1.name == '1'
    assert region_2.name == '2'

    assert np.array_equal(region_1, np.ones((1, 5), dtype=np.bool))
    temp = np.ones((2, 2), dtype=np.bool)
    temp[-1, -1] = False
    assert np.array_equal(region_2, temp)

    assert np.array_equal(region_1[Axes.Y.value], [0])
    assert np.array_equal(region_1[Axes.X.value], [0, 1, 2, 3, 4])

    assert np.array_equal(region_2[Axes.Y.value], [3, 4])
    assert np.array_equal(region_2[Axes.X.value], [3, 4])

    assert np.array_equal(region_1[Coordinates.Y.value],
                          physical_ticks[Coordinates.Y][0:1])
    assert np.array_equal(region_1[Coordinates.X.value],
                          physical_ticks[Coordinates.X][0:5])

    assert np.array_equal(region_2[Coordinates.Y.value],
                          physical_ticks[Coordinates.Y][3:5])
    assert np.array_equal(region_2[Coordinates.X.value],
                          physical_ticks[Coordinates.X][3:5])


def test_save_load():
    label_image = np.zeros((5, 5), dtype=np.int32)
    label_image[0] = 1
    label_image[3:5, 3:5] = 2
    label_image[-1, -1] = 0

    physical_ticks = {Coordinates.Y: [1.2, 2.4, 3.6, 4.8, 6.0],
                      Coordinates.X: [7.2, 8.4, 9.6, 10.8, 12]}

    masks = SegmentationMaskCollection.from_label_image(label_image,
                                                        physical_ticks)

    path = 'data.tgz'
    try:
        masks.save(path)
        masks2 = SegmentationMaskCollection.from_disk(path)
        for m, m2 in zip(masks, masks2):
            assert np.array_equal(m, m2)
    finally:
        os.remove(path)



from starfish.core.experiment.builder import (   # noqa: F401
    build_image,
    build_irregular_image,
    inplace,
    TileIdentifier,
    write_experiment_json,
    write_irregular_experiment_json,
)
from starfish.core.experiment.builder.providers import FetchedTile, TileFetcher  # noqa: F401



from functools import partial
from typing import Optional, Union

import numpy as np
import xarray as xr

from starfish.core.imagestack.imagestack import ImageStack
from starfish.core.types import Number
from starfish.core.util import click
from ._base import FilterAlgorithmBase
from .util import determine_axes_to_group_by


class ClipValueToZero(FilterAlgorithmBase):
    """
    Image clipping filter that clips values below a minimum value and above a
    maximum value. The filter then subtracts the minimum value from the
    clipped image.

    By default, the min and max values are set to 0.0 and None respectively,
    which will result in the filter doing nothing.

    This is a wrapper for :py:func:`numpy.clip`.

    Parameters
    ----------
    v_min : float
        Values below v_min are set to v_min. v_min is then subtracted from the
        entire image (default 0)
    v_max : Optional[Number]
        Values above v_max are set to v_max (default None)
    is_volume : bool
        If True, 3d (z, y, x) volumes will be filtered. By default, filter 2D
        (y, x) tiles
    """

    def __init__(self,
                 v_min: float = 0.0,
                 v_max: Optional[Number] = None,
                 is_volume: bool = False) -> None:
        self.v_min = v_min
        self.v_max = v_max
        self.is_volume = is_volume

    _DEFAULT_TESTING_PARAMETERS = {"v_min": 0.0, "v_max": None}

    @staticmethod
    def _clip_value_to_zero(image: Union[xr.DataArray, np.ndarray],
                            v_min: float,
                            v_max: Optional[Number]) -> np.ndarray:
        return image.clip(min=v_min, max=v_max) - np.float32(v_min)

    def run(self,
            stack: ImageStack,
            in_place: bool = False,
            verbose: bool = False,
            n_processes: Optional[int] = None,
            *args) -> ImageStack:
        """Perform filtering of an image stack

        Parameters
        ----------
        stack : ImageStack
            Stack to be filtered.
        in_place : bool
            if True, process ImageStack in-place, otherwise return a new stack
        verbose : bool
            if True, report on filtering progress (default = False)
        n_processes : Optional[int]
            Number of parallel processes to devote to applying the filter.
            If None, defaults to the result of os.cpu_count(). (default None)

        Returns
        -------
        ImageStack :
            If in-place is False, return the results of filter as a new stack.
            Otherwise return the original stack.

        """
        group_by = determine_axes_to_group_by(self.is_volume)
        clip_value_to_zero = partial(
            self._clip_value_to_zero,
            v_min=self.v_min, v_max=self.v_max,
        )
        result = stack.apply(
            clip_value_to_zero,
            group_by=group_by, verbose=verbose,
            in_place=in_place, n_processes=n_processes
        )
        return result

    @staticmethod
    @click.command("ClipValueToZero")
    @click.option(
        "--v-min", default=0.0, type=float,
        help=("clip intensities below this value and subtract this value "
              "from the image"))
    @click.option(
        "--v-max", default=None, type=float,
        help="clip intensities above this value")
    @click.option(
        "--is-volume", is_flag=True, help="filter 3D volumes")
    @click.pass_context
    def _cli(ctx, v_min, v_max, is_volume):
        ctx.obj["component"]._cli_run(ctx,
                                      ClipValueToZero(v_min, v_max, is_volume))



import inspect
import warnings
from typing import (
    Any,
    Callable,
    cast,
    Mapping,
    MutableMapping,
    MutableSequence,
    Sequence,
    Set,
    Type,
)

from starfish.core.pipeline.algorithmbase import AlgorithmBase
from .errors import (
    ConstructorError,
    ConstructorExtraParameterWarning,
    ExecutionError,
    RunInsufficientParametersError,
    TypeInferenceError,
)
from .filesystem import FileProvider, TypedFileProvider


class Runnable:
    """Runnable represents a single invocation of a pipeline component, with a specific algorithm
    implementation.  For arguments to the algorithm's constructor and run method, it can accept
    :py:class:`~starfish.recipe.filesystem.FileProvider` objects, which represent a file path or
    url.  For arguments to the algorithm's run method, it can accept the results of other Runnables.

    One can compose any starfish pipeline run using a directed acyclic graph of Runnables objects.
    """
    def __init__(
            self,
            algorithm_cls: Type,
            *inputs,
            **algorithm_options
    ) -> None:
        self._pipeline_component_cls: Type = \
            cast(AlgorithmBase, algorithm_cls).get_pipeline_component_class()
        self._algorithm_cls: Type = algorithm_cls
        self._raw_inputs = inputs
        self._raw_algorithm_options = algorithm_options

        # retrieve the actual __init__ method
        signature = Runnable._get_actual_method_signature(
            self._algorithm_cls.__init__)
        formatted_algorithm_options = self._format_algorithm_constructor_arguments(
            signature, self._raw_algorithm_options, self.__str__)

        try:
            self._algorithm_instance: AlgorithmBase = self._algorithm_cls(
                **formatted_algorithm_options)
        except Exception as ex:
            raise ConstructorError(f"Error instantiating the algorithm for {str(self)}") from ex

        # retrieve the actual run method
        signature = Runnable._get_actual_method_signature(
            self._algorithm_instance.run)  # type: ignore
        self._inputs = self._format_run_arguments(signature, self._raw_inputs, self.__str__)

    @staticmethod
    def _get_actual_method_signature(run_method: Callable) -> inspect.Signature:
        if hasattr(run_method, "__closure__"):
            # it's a closure, probably because of AlgorithmBaseType.run_with_logging.  Unwrap to
            # find the underlying method.
            closure = run_method.__closure__  # type: ignore
            if closure is not None:
                run_method = closure[0].cell_contents

        return inspect.signature(run_method)

    @staticmethod
    def _format_algorithm_constructor_arguments(
            constructor_signature: inspect.Signature,
            algorithm_options: Mapping[str, Any],
            str_callable: Callable[[], str],
    ) -> Mapping[str, Any]:
        """Given the constructor's signature and a mapping of keyword argument names to values,
        format them such that the constructor can be invoked.

        Some parameters may be :py:class:`starfish.recipe.filesystem.FileProvider` instances.  Use
        the type hints in the constructor's signature to identify the expected file type, and load
        them into memory accordingly.

        Parameters
        ----------
        constructor_signature : inspect.Signature
            The signature for the constructor.
        algorithm_options : Mapping[str, Any]
            The parameters for the constructor, as provided to the Runnable.
        str_callable : Callable[[], str]
            A callable that can be invoked to provide a user-friendly representation of the
            Runnable, in case any errors or warnings are generated.

        Returns
        -------
        Mapping[str, Any] : The parameters for the constructor, ready to be passed into the
        constructor.
        """
        parameters = constructor_signature.parameters

        formatted_algorithm_options: MutableMapping[str, Any] = {}
        for algorithm_option_name, algorithm_option_value in algorithm_options.items():
            if isinstance(algorithm_option_value, Runnable):
                raise RuntimeError("Runnable's constructors cannot depend on another runnable")

            try:
                option_class = parameters[algorithm_option_name].annotation
            except KeyError:
                warnings.warn(
                    f"Constructor for {str_callable()} does not have an explicitly typed "
                    + f"parameter {algorithm_option_name}.",
                    category=ConstructorExtraParameterWarning,
                )
                continue

            if isinstance(algorithm_option_value, FileProvider):
                try:
                    provider = TypedFileProvider(algorithm_option_value, option_class)
                except TypeError as ex:
                    raise TypeInferenceError(
                        f"Error inferring the types for the parameters to the algorithm's"
                        + f" constructor for {str_callable()}") from ex
                formatted_algorithm_options[algorithm_option_name] = provider.load()
            else:
                formatted_algorithm_options[algorithm_option_name] = algorithm_option_value

        return formatted_algorithm_options

    @staticmethod
    def _format_run_arguments(
            run_signature: inspect.Signature,
            inputs: Sequence,
            str_callable: Callable[[], str],
    ) -> Sequence:
        """Given the run method's signature and a sequence of parameters, format them such that the
        run method can be invoked.

        Some parameters may be :py:class:`starfish.recipe.filesystem.FileProvider` instances.  Use
        the type hints in the run method's signature to identify the expected file type, and load
        them into memory accordingly.

        Parameters that are the outputs of other Runnables are not resolved to their values until
        the run method is invoked.  Therefore, the sequence of parameters returned may include
        the dependent Runnable objects.

        Parameters
        ----------
        run_signature : inspect.Signature
            The signature for the run method.
        inputs : Sequence
            The parameters for the run method, as provided to the Runnable.
        str_callable : Callable[[], str]
            A callable that can be invoked to provide a user-friendly representation of the
            Runnable, in case any errors or warnings are generated.

        Returns
        -------
        Sequence : The parameters for the run method, ready to be passed into the constructor,
                   except for dependent Runnables, which are resolved later.
        """
        formatted_inputs: MutableSequence = []

        keys_iter = iter(run_signature.parameters.keys())
        inputs_iter = iter(inputs)

        # first parameter to the run method should be "self"
        assert next(keys_iter) == "self"

        # match up the parameters as best as we can.
        for _input, key in zip(inputs_iter, keys_iter):
            if isinstance(_input, FileProvider):
                annotation = run_signature.parameters[key].annotation
                try:
                    provider = TypedFileProvider(_input, annotation)
                except TypeError as ex:
                    raise TypeInferenceError(
                        f"Error inferring the types for the parameters to the algorithm's"
                        + f" run method for {str_callable()}") from ex
                formatted_inputs.append(provider)
            else:
                formatted_inputs.append(_input)

        # are there any parameters left in the signature?  if so, they must have default values
        # because we don't have values.
        no_default = inspect._empty  # type: ignore

        for key in keys_iter:
            if (run_signature.parameters[key].default == no_default
                    and run_signature.parameters[key].kind not in (
                        inspect.Parameter.VAR_POSITIONAL,
                        inspect.Parameter.VAR_KEYWORD,
            )):
                raise RunInsufficientParametersError(f"No value for parameter {key}")

        return formatted_inputs

    @property
    def runnable_dependencies(self) -> Set["Runnable"]:
        """Retrieves a set of Runnables that this Runnable depends on."""
        return set(runnable for runnable in self._inputs if isinstance(runnable, Runnable))

    def run(self, previous_results: Mapping["Runnable", Any]) -> Any:
        """Invoke the run method.  Results for dependent Runnables are retrieved from the
        `previous_results` mapping.

        Parameters
        ----------
        previous_results : Mapping[Runnable, Any]
            The results calculated thus far in an execution run.

        Returns
        -------
        The result from invoking the run method.
        """
        inputs = list()
        for _input in self._inputs:
            if isinstance(_input, Runnable):
                inputs.append(previous_results[_input])
            elif isinstance(_input, TypedFileProvider):
                inputs.append(_input.load())
            else:
                inputs.append(_input)
        try:
            return self._algorithm_instance.run(*inputs)  # type: ignore
        except Exception as ex:
            raise ExecutionError(f"Error running the algorithm for {str(self)}") from ex

    def __str__(self):
        inputs_arr = [""]
        inputs_arr.extend([str(raw_input) for raw_input in self._raw_inputs])
        algorithm_options_arr = [""]
        algorithm_options_arr.extend([
            f"{algorithm_option_name}={str(algorithm_option_value)}"
            for algorithm_option_name, algorithm_option_value in
            self._raw_algorithm_options.items()
        ])

        inputs_str = ", ".join(inputs_arr)
        algorithm_options_str = ", ".join(algorithm_options_arr)

        return (f"compute("
                + f"{self._pipeline_component_cls.__name__}.{self._algorithm_cls.__name__}"
                + f"{inputs_str}"
                + f"{algorithm_options_str})")



from slicedimage.io import resolve_path_or_url

from starfish.core.config import StarfishConfig
from starfish.core.starfish import starfish
from starfish.core.util import click
from .recipe import Recipe


@starfish.command("recipe")
@click.option("--recipe", required=True, type=str, metavar="RECIPE_PATH_OR_URL")
@click.option(
    "--input", type=str, multiple=True, metavar="INPUT_FILE_PATH_OR_URL",
    help="input file paths or urls to map to the recipe input parameters")
@click.option(
    "--output", type=str, multiple=True, metavar="OUTPUT_FILE_PATH",
    help="output file paths to write recipe outputs to")
@click.pass_context
def run_recipe(ctx, recipe, input, output):
    """Runs a recipe with a given set of inputs and outputs."""
    config = StarfishConfig()

    backend, relativeurl, _ = resolve_path_or_url(
        recipe, backend_config=config.slicedimage)
    with backend.read_contextmanager(relativeurl) as fh:
        recipe_str = fh.read()

    recipe_obj = Recipe(recipe_str, input, output)
    recipe_obj.run_and_save()



# constants
from starfish.core.types import (  # noqa: F401
    Axes,
    Clip,
    Coordinates,
    CORE_DEPENDENCIES,
    Features,
    LOG,
    OverlapStrategy,
    PHYSICAL_COORDINATE_DIMENSION,
    PhysicalCoordinateTypes,
    STARFISH_EXTRAS_KEY,
    TransformType,
)
from starfish.core.types import CoordinateValue, Number  # noqa: F401



from typing import (
    AbstractSet,
    Any,
    Mapping,
    MutableMapping,
    MutableSequence,
    Optional,
    Sequence,
    Set,
)

import numpy as np

from starfish import image, spots
from starfish.types import Axes, Coordinates
from .filesystem import FileProvider, FileTypes
from .runnable import Runnable


class ExecutionComplete(Exception):
    """Raised by :py:class:`Execution` when it is complete.  We don't rely on catching StopIteration
    because some underlying library may have raised that instead."""
    pass


class _Execution:
    """Encompasses the state of a single execution of a recipe."""
    def __init__(
            self,
            runnable_sequence: Sequence[Runnable],
            output_runnables: Sequence[Runnable],
            output_paths: Sequence[str],
    ) -> None:
        self._runnable_sequence = iter(runnable_sequence)
        self._output_runnables = output_runnables
        self._output_paths = output_paths

        # build a map between each runnable and its dependents.  each time a runnable completes, we
        # go through each of its dependencies to see if its results are still needed.
        runnable_dependents: MutableMapping[Runnable, Set[Runnable]] = dict()
        for output_runnable in self._output_runnables:
            _Execution._build_graph(output_runnable, runnable_dependents)
        self.runnable_dependents: Mapping[Runnable, AbstractSet[Runnable]] = runnable_dependents

        # completed results
        self._completed_runnables: Set[Runnable] = set()
        self._completed_results: MutableMapping[Runnable, Any] = dict()

    def _run_one_tick(self) -> None:
        """Run one tick of the execution graph.  Raises StopIteration if it's done."""
        try:
            runnable = next(self._runnable_sequence)
        except StopIteration as ex:
            raise ExecutionComplete from ex

        result = runnable.run(self._completed_results)

        # record what's been done.
        self._completed_runnables.add(runnable)
        self._completed_results[runnable] = result

        # examine all the dependencies, and discard the results if no one else needs it.
        for dependency in runnable.runnable_dependencies:
            if dependency in self._output_runnables:
                # it's required by the outputs, so preserve this.
                continue

            for dependent in self.runnable_dependents[dependency]:
                if dependent not in self._completed_runnables:
                    # someone still needs this runnable's result.
                    break
            else:
                # every dependent is complete.  drop the result.
                del self._completed_results[dependency]

    def _save(self) -> None:
        for runnable, output_path in zip(self._output_runnables, self._output_paths):
            # get the result
            result = self._completed_results[runnable]

            filetype = FileTypes.resolve_by_instance(result)
            filetype.save(result, output_path)

    def run_and_save(self) -> None:
        """Run execution graph to completion and save the results to disk."""
        while True:
            try:
                self._run_one_tick()
            except ExecutionComplete:
                break

        self._save()

    @staticmethod
    def _build_graph(
            runnable: Runnable,
            runnable_dependents: MutableMapping[Runnable, Set[Runnable]],
            seen_runnables: Optional[Set[Runnable]]=None,
    ) -> None:
        if seen_runnables is None:
            seen_runnables = set()

        if runnable in seen_runnables:
            return
        seen_runnables.add(runnable)

        for dependency in runnable.runnable_dependencies:
            # mark ourselves a dependent of each of our dependencies.
            if dependency not in runnable_dependents:
                runnable_dependents[dependency] = set()
            runnable_dependents[dependency].add(runnable)
            _Execution._build_graph(dependency, runnable_dependents, seen_runnables)


class OrderedSequence:
    def __init__(self) -> None:
        self._sequence: MutableSequence[Runnable] = list()

    def __call__(self, *args, **kwargs):
        result = Runnable(*args, **kwargs)
        self._sequence.append(result)
        return result

    @property
    def sequence(self) -> Sequence[Runnable]:
        return self._sequence


class Recipe:
    """Recipes are snippets of python code that describe an image processing pipeline.  Executing a
    recipe requires pairing the recipe code with locations of input files and the locations to write
    outputs to.

    A few additional variables are injected into the recipe's scope:

    ====================  ==========================================================================
     Variable name         Description
    --------------------  --------------------------------------------------------------------------
     file_inputs           A 0-based array with the locations of input files provided for a recipe's
                           execution.
     file_outputs          Any objects written to this 0-based array will be written out to disk at
                           the end of a recipe's execution.
     np                    A reference to the numpy library.
     Axes                  A reference to the :py:class:`~starfish.types.Axes` enum.
     Coordinates           A reference to the :py:class:`~starfish.types.Coordinates` enum.
     ApplyTransform        A reference to :py:class:`~starfish.image.ApplyTransform`.
     Filter                A reference to :py:class:`~starfish.image.Filter`.
     LearnTransform        A reference to :py:class:`~starfish.image.LearnTransform`.
     Segment               A reference to :py:class:`~starfish.image.Segment`.
     Decode                A reference to :py:class:`~starfish.spots.Decode`.
     DetectPixels          A reference to :py:class:`~starfish.spots.DetectPixels`.
     DetectSpots           A reference to :py:class:`~starfish.spots.DetectSpots`.
     AssignTargets         A reference to :py:class:`~starfish.spots.AssignTargets`.
    ====================  ==========================================================================

    Recipes also have access to a ``compute`` method:
    :code:`compute(algorithm_class, parameters_to_run_method...,
    constructor_argument_name=constructor_argument_value....)`

    The first parameter is the algorithm class that implements a pipeline algorithm.  It should
    subclass :py:class:`~starfish.pipeline.AlgorithmBase`.

    The subsequent positional parameters are the parameters passed to the ``run()`` method of the
    algorithm class.

    The keyword parameters (parameters where a name and a value are provided) are the parameters
    passed to the algorithm class's constructor.

    Parameters can include input files that are passed in as a member of the ``file_inputs`` array.
    It can also include the output of previous ``compute()`` method calls.

    The return value of ``compute()`` method calls is a future, i.e., a promise for the data.  It
    should not be assumed that the resulting data will be available during the recipe's scope.
    """
    def __init__(
            self,
            recipe_str: str,
            input_paths_or_urls: Sequence[str],
            output_paths: Sequence[str],
    ):
        ordered_sequence = OrderedSequence()
        file_outputs: MutableMapping[int, Runnable] = {}
        recipe_scope = {
            "file_inputs": [
                FileProvider(input_path_or_url)
                for input_path_or_url in input_paths_or_urls
            ],
            "compute": ordered_sequence,
            "file_outputs": file_outputs,
        }

        # inject numpy.
        recipe_scope['np'] = np

        component: Any

        # inject some useful starfish classes.
        for component in (Axes, Coordinates):
            recipe_scope[component.__name__] = component

        # inject all the pipeline components.
        for component in (
                image.ApplyTransform, image.Filter, image.LearnTransform, image.Segment):
            recipe_scope[component.__name__] = component
        for component in (
                spots.Decode, spots.DetectPixels, spots.DetectSpots, spots.AssignTargets):
            recipe_scope[component.__name__] = component

        ast = compile(recipe_str, "<string>", "exec")
        exec(ast, recipe_scope)

        assert len(file_outputs) == len(output_paths), \
            "Recipe generates more outputs than output paths provided!"

        # verify that the outputs are sequential.
        ordered_outputs: MutableSequence[Runnable] = list()
        for ix in range(len(file_outputs)):
            assert ix in file_outputs, \
                f"file_outputs[{ix}] is not set"
            assert isinstance(file_outputs[ix], Runnable), \
                f"file_outputs[{ix}] is not the result of a compute(..)"
            ordered_outputs.append(file_outputs[ix])

        self._runnable_order = ordered_sequence.sequence
        self._outputs: Sequence[Runnable] = ordered_outputs
        self._output_paths = output_paths

    def _execution(self) -> _Execution:
        return _Execution(self._runnable_order, self._outputs, self._output_paths)

    def run_and_save(self):
        """Run recipe to completion and save the results to disk."""
        execution = self._execution()
        execution.run_and_save()



import functools
import json
import os
from dataclasses import astuple, dataclass
from pathlib import Path
from typing import (
    Any,
    BinaryIO,
    Callable,
    Dict,
    Iterable,
    Mapping,
    MutableMapping,
    MutableSequence,
    MutableSet,
    Optional,
    Sequence,
    Union,
)

from slicedimage import (
    Collection,
    ImageFormat,
    Tile,
    TileSet,
    Writer,
)

from starfish import FieldOfView
from starfish.core.codebook.codebook import Codebook
from starfish.core.experiment.builder.orderediterator import join_axes_labels, ordered_iterator
from starfish.core.experiment.version import CURRENT_VERSION
from starfish.core.types import Axes, Coordinates
from .defaultproviders import RandomNoiseTile, tile_fetcher_factory
from .providers import FetchedTile, TileFetcher


DEFAULT_DIMENSION_ORDER = (Axes.ZPLANE, Axes.ROUND, Axes.CH)


@dataclass(eq=True, order=True, frozen=True)
class TileIdentifier:
    """Data class for encapsulating the location of a tile in a 6D tensor (fov, round, ch, zplane,
    y, and x)."""
    fov_id: int
    round_label: int
    ch_label: int
    zplane_label: int


def _tile_opener(toc_path: Path, tile: Tile, file_ext: str) -> BinaryIO:
    base = toc_path.parent / toc_path.stem
    return open(
        f"{os.fspath(base)}-Z{tile.indices[Axes.ZPLANE]}-"
        f"R{tile.indices[Axes.ROUND]}-C{tile.indices[Axes.CH]}.{file_ext}",
        "wb")


def _fov_path_generator(parent_toc_path: Path, toc_name: str) -> Path:
    return parent_toc_path.parent / "{}-{}.json".format(parent_toc_path.stem, toc_name)


def build_irregular_image(
        tile_identifiers: Iterable[TileIdentifier],
        image_fetcher: TileFetcher,
        default_shape: Optional[Mapping[Axes, int]] = None,
) -> Collection:
    """
    Build and returns an image set that can potentially be irregular (i.e., the cardinality of the
    dimensions are not always consistent).  It can also build a regular image.

    Parameters
    ----------
    tile_identifiers : Iterable[TileIdentifier]
        Iterable of all the TileIdentifier that are valid in the image.
    image_fetcher : TileFetcher
        Instance of TileFetcher that provides the data for the tile.
    default_shape : Optional[Tuple[int, int]]
        Default shape of the individual tiles in this image set.

    Returns
    -------
    The slicedimage collection representing the image.
    """
    def reducer_to_sets(
            accumulated: Sequence[MutableSet[int]], update: TileIdentifier,
    ) -> Sequence[MutableSet[int]]:
        """Reduces to a list of sets of tile identifiers, in the order of FOV, round, ch, and
        zplane."""
        result: MutableSequence[MutableSet[int]] = list()
        for accumulated_elem, update_elem in zip(accumulated, astuple(update)):
            accumulated_elem.add(update_elem)
            result.append(accumulated_elem)
        return result
    initial_value: Sequence[MutableSet[int]] = tuple(set() for _ in range(4))

    fovs, rounds, chs, zplanes = functools.reduce(
        reducer_to_sets, tile_identifiers, initial_value)

    collection = Collection()
    for expected_fov in fovs:
        fov_images = TileSet(
            [
                Coordinates.X,
                Coordinates.Y,
                Coordinates.Z,
                Axes.ZPLANE,
                Axes.ROUND,
                Axes.CH,
                Axes.X,
                Axes.Y,
            ],
            {Axes.ROUND: len(rounds), Axes.CH: len(chs), Axes.ZPLANE: len(zplanes)},
            default_shape,
            ImageFormat.TIFF,
        )

        for tile_identifier in tile_identifiers:
            current_fov, current_round, current_ch, current_zplane = astuple(tile_identifier)
            # filter out the fovs that are not the one we are currently processing
            if expected_fov != current_fov:
                continue
            image = image_fetcher.get_tile(
                current_fov,
                current_round,
                current_ch,
                current_zplane
            )
            tile = Tile(
                image.coordinates,
                {
                    Axes.ZPLANE: current_zplane,
                    Axes.ROUND: current_round,
                    Axes.CH: current_ch,
                },
                image.shape,
                extras=image.extras,
            )
            tile.set_numpy_array_future(image.tile_data)
            # Astute readers might wonder why we set this variable.  This is to support in-place
            # experiment construction.  We monkey-patch slicedimage's Tile class such that checksum
            # computation is done by finding the FetchedTile object, which allows us to calculate
            # the checksum of the original file.
            tile.provider = image
            fov_images.add_tile(tile)
        collection.add_partition("fov_{:03}".format(expected_fov), fov_images)
    return collection


def build_image(
        fovs: Sequence[int],
        rounds: Sequence[int],
        chs: Sequence[int],
        zplanes: Sequence[int],
        image_fetcher: TileFetcher,
        default_shape: Optional[Mapping[Axes, int]] = None,
        axes_order: Sequence[Axes] = DEFAULT_DIMENSION_ORDER,
) -> Collection:
    """
    Build and returns an image set with the following characteristics:

    Parameters
    ----------
    fovs : Sequence[int]
        Sequence of field of view ids in this image set.
    rounds : Sequence[int]
        Sequence of the round numbers in this image set.
    chs : Sequence[int]
        Sequence of the ch numbers in this image set.
    zplanes : Sequence[int]
        Sequence of the zplane numbers in this image set.
    image_fetcher : TileFetcher
        Instance of TileFetcher that provides the data for the tile.
    default_shape : Optional[Tuple[int, int]]
        Default shape of the individual tiles in this image set.
    axes_order : Sequence[Axes]
        Ordering for which axes vary, in order of the slowest changing axis to the fastest.  For
        instance, if the order is (ROUND, Z, CH) and each dimension has size 2, then the sequence
        is:
          (ROUND=0, CH=0, Z=0)
          (ROUND=0, CH=1, Z=0)
          (ROUND=0, CH=0, Z=1)
          (ROUND=0, CH=1, Z=1)
          (ROUND=1, CH=0, Z=0)
          (ROUND=1, CH=1, Z=0)
          (ROUND=1, CH=0, Z=1)
          (ROUND=1, CH=1, Z=1)
        (default = (Axes.Z, Axes.ROUND, Axes.CH))

    Returns
    -------
    The slicedimage collection representing the image.
    """
    axes_sizes = join_axes_labels(
        axes_order, rounds=rounds, chs=chs, zplanes=zplanes)
    tile_identifiers = [
        TileIdentifier(fov_id, selector[Axes.ROUND], selector[Axes.CH], selector[Axes.ZPLANE])
        for fov_id in fovs
        for selector in ordered_iterator(axes_sizes)
    ]

    return build_irregular_image(tile_identifiers, image_fetcher, default_shape)


def write_irregular_experiment_json(
        path: str,
        tile_format: ImageFormat,
        *,
        image_tile_identifiers: Mapping[str, Iterable[TileIdentifier]],
        tile_fetchers: Mapping[str, TileFetcher],
        postprocess_func: Optional[Callable[[dict], dict]]=None,
        default_shape: Optional[Mapping[Axes, int]]=None,
        fov_path_generator: Callable[[Path, str], Path] = None,
        tile_opener: Optional[Callable[[Path, Tile, str], BinaryIO]] = None,
) -> None:
    """
    Build and returns a top-level experiment description with the following characteristics:

    Parameters
    ----------
    path : str
        Directory to write the files to.
    tile_format : ImageFormat
        File format to write the tiles as.
    image_tile_identifiers : Mapping[str, Iterable[TileIdentifier]]
        Dictionary mapping the image type to an iterable of TileIdentifiers.
    tile_fetchers : Mapping[str, TileFetcher]
        Dictionary mapping the image type to a TileFetcher.
    postprocess_func : Optional[Callable[[dict], dict]]
        If provided, this is called with the experiment document for any postprocessing.
        An example of this would be to add something to one of the top-level extras field.
        The callable should return what is to be written as the experiment document.
    default_shape : Optional[Tuple[int, int]] (default = None)
        Default shape for the tiles in this experiment.
    fov_path_generator : Optional[Callable[[Path, str], Path]]
        Generates the path for a FOV's json file.  If one is not provided, the default generates
        the FOV's json file at the same level as the top-level json file for an image.    If this is
        not provided, a reasonable default will be provided.
    tile_opener : Optional[Callable[[Path, Tile, str], BinaryIO]]
        Callable that gets invoked with the following arguments: 1. the directory of the experiment
        that is being constructed, 2. the tile that is being written, and 3. the file extension
        that the tile should be written with.  The callable is expected to return an open file
        handle.  If this is not provided, a reasonable default will be provided.
    """
    if postprocess_func is None:
        postprocess_func = lambda doc: doc
    if fov_path_generator is None:
        fov_path_generator = _fov_path_generator
    if tile_opener is None:
        tile_opener = _tile_opener

    experiment_doc: Dict[str, Any] = {
        'version': str(CURRENT_VERSION),
        'images': {},
        'extras': {},
    }
    for image_type, tile_identifiers in image_tile_identifiers.items():
        tile_fetcher = tile_fetchers[image_type]

        image = build_irregular_image(tile_identifiers, tile_fetcher, default_shape)

        Writer.write_to_path(
            image,
            os.path.join(path, f"{image_type}.json"),
            pretty=True,
            partition_path_generator=fov_path_generator,
            tile_opener=tile_opener,
            tile_format=tile_format,
        )
        experiment_doc['images'][image_type] = f"{image_type}.json"

    experiment_doc["codebook"] = "codebook.json"
    codebook_array = [
        {
            "codeword": [
                {"r": 0, "c": 0, "v": 1},
            ],
            "target": "PLEASE_REPLACE_ME"
        },
    ]
    codebook = Codebook.from_code_array(codebook_array)
    codebook_json_filename = "codebook.json"
    codebook.to_json(os.path.join(path, codebook_json_filename))

    experiment_doc = postprocess_func(experiment_doc)

    with open(os.path.join(path, "experiment.json"), "w") as fh:
        json.dump(experiment_doc, fh, indent=4)


def write_experiment_json(
        path: str,
        fov_count: int,
        tile_format: ImageFormat,
        *,
        primary_image_dimensions: Mapping[Union[str, Axes], int],
        aux_name_to_dimensions: Mapping[str, Mapping[Union[str, Axes], int]],
        primary_tile_fetcher: Optional[TileFetcher]=None,
        aux_tile_fetcher: Optional[Mapping[str, TileFetcher]]=None,
        postprocess_func: Optional[Callable[[dict], dict]]=None,
        default_shape: Optional[Mapping[Axes, int]]=None,
        dimension_order: Sequence[Axes]=(Axes.ZPLANE, Axes.ROUND, Axes.CH),
        fov_path_generator: Optional[Callable[[Path, str], Path]] = None,
        tile_opener: Optional[Callable[[Path, Tile, str], BinaryIO]] = None,
) -> None:
    """
    Build and returns a top-level experiment description with the following characteristics:

    Parameters
    ----------
    path : str
        Directory to write the files to.
    fov_count : int
        Number of fields of view in this experiment.
    tile_format : ImageFormat
        File format to write the tiles as.
    primary_image_dimensions : Mapping[Union[str, Axes], int]
        Dictionary mapping dimension name to dimension size for the primary image.
    aux_name_to_dimensions : Mapping[str, Mapping[Union[str, Axes], int]]
        Dictionary mapping the auxiliary image type to dictionaries, which map from dimension name
        to dimension size.
    primary_tile_fetcher : Optional[TileFetcher]
        TileFetcher for primary images.  Set this if you want specific image data to be set for the
        primary images.  If not provided, the image data is set to random noise via
        :class:`RandomNoiseTileFetcher`.
    aux_tile_fetcher : Optional[Mapping[str, TileFetcher]]
        TileFetchers for auxiliary images.  Set this if you want specific image data to be set for
        one or more aux image types.  If not provided for any given aux image, the image data is
        set to random noise via :class:`RandomNoiseTileFetcher`.
    postprocess_func : Optional[Callable[[dict], dict]]
        If provided, this is called with the experiment document for any postprocessing.
        An example of this would be to add something to one of the top-level extras field.
        The callable should return what is to be written as the experiment document.
    default_shape : Optional[Tuple[int, int]] (default = None)
        Default shape for the tiles in this experiment.
    dimension_order : Sequence[Axes]
        Ordering for which dimensions vary, in order of the slowest changing dimension to the
        fastest.  For instance, if the order is (ROUND, Z, CH) and each dimension has size 2, then
        the sequence is:
          (ROUND=0, CH=0, Z=0)
          (ROUND=0, CH=1, Z=0)
          (ROUND=0, CH=0, Z=1)
          (ROUND=0, CH=1, Z=1)
          (ROUND=1, CH=0, Z=0)
          (ROUND=1, CH=1, Z=0)
          (ROUND=1, CH=0, Z=1)
          (ROUND=1, CH=1, Z=1)
        (default = (Axes.Z, Axes.ROUND, Axes.CH))
    fov_path_generator : Callable[[Path, str], Path]
        Generates the path for a FOV's json file.  If one is not provided, the default generates
        the FOV's json file at the same level as the top-level json file for an image.  If this is
        not provided, a reasonable default will be provided.
    tile_opener : Callable[[Path, Tile, str], BinaryIO]
        Callable that gets invoked with the following arguments: 1. the directory of the experiment
        that is being constructed, 2. the tile that is being written, and 3. the file extension
        that the tile should be written with.  The callable is expected to return an open file
        handle.  If this is not provided, a reasonable default will be provided.
    """
    all_tile_fetcher: MutableMapping[str, TileFetcher] = {}
    if aux_tile_fetcher is not None:
        all_tile_fetcher.update(aux_tile_fetcher)
    if primary_tile_fetcher is not None:
        all_tile_fetcher[FieldOfView.PRIMARY_IMAGES] = primary_tile_fetcher

    image_tile_identifiers: MutableMapping[str, Iterable[TileIdentifier]] = dict()
    image_tile_fetchers: MutableMapping[str, TileFetcher] = dict()

    dimension_cardinality_of_images = {FieldOfView.PRIMARY_IMAGES: primary_image_dimensions}
    for image_type, image_dimension in aux_name_to_dimensions.items():
        dimension_cardinality_of_images[image_type] = image_dimension

    for image_type, dimension_cardinality_of_image in dimension_cardinality_of_images.items():
        axes_sizes = join_axes_labels(
            dimension_order,
            rounds=range(dimension_cardinality_of_image[Axes.ROUND]),
            chs=range(dimension_cardinality_of_image[Axes.CH]),
            zplanes=range(dimension_cardinality_of_image[Axes.ZPLANE]),
        )
        image_tile_identifiers[image_type] = [
            TileIdentifier(fov_id, selector[Axes.ROUND], selector[Axes.CH], selector[Axes.ZPLANE])
            for fov_id in range(fov_count)
            for selector in ordered_iterator(axes_sizes)
        ]

        image_tile_fetchers[image_type] = all_tile_fetcher.get(
            image_type, tile_fetcher_factory(RandomNoiseTile))

    return write_irregular_experiment_json(
        path, tile_format,
        image_tile_identifiers=image_tile_identifiers,
        tile_fetchers=image_tile_fetchers,
        postprocess_func=postprocess_func,
        default_shape=default_shape,
        fov_path_generator=fov_path_generator,
        tile_opener=tile_opener,
    )



import re

from starfish.core.experiment.experiment import Experiment
from starfish.core.imagestack.imagestack import ImageStack
from starfish.core.util.indirectfile._base import ConversionFormula


CRE = re.compile("@(?P<path>.+)\[(?P<fov>[^\[\]]+)\]\[(?P<image_type>[^\[\]]+)\]")  # noqa: W605


class GetImageStackFromExperiment(ConversionFormula[ImageStack]):
    def applicable(self, input_parameter: str) -> bool:
        return CRE.match(input_parameter) is not None

    def load(self, input_parameter: str) -> ImageStack:
        mo = CRE.match(input_parameter)
        assert mo is not None
        experiment = Experiment.from_json(mo.group("path"))
        fov = experiment[mo.group("fov")]
        return fov.get_image(mo.group("image_type"))


class GetImageStack(ConversionFormula[ImageStack]):
    def applicable(self, input_parameter: str) -> bool:
        return not CRE.match(input_parameter)

    def load(self, input_parameter: str) -> ImageStack:
        return ImageStack.from_path_or_url(input_parameter)



from ._base import (
    ConversionFormula,
    convert,
    NoApplicableConversionFormulaError,
    NoSuccessfulConversionFormulaError,
)
from ._codebook import GetCodebook, GetCodebookFromExperiment
from ._imagestack import GetImageStack, GetImageStackFromExperiment



import __main__
import warnings
from collections import OrderedDict
from typing import Iterable, List, Optional, Set, Tuple, Union

import numpy as np

from starfish.core.imagestack.imagestack import ImageStack
from starfish.core.intensity_table.intensity_table import IntensityTable
from starfish.core.types import Axes, Features

try:
    with warnings.catch_warnings():
        warnings.filterwarnings("ignore", module="vispy.visuals.isocurve", lineno=22)
        from napari import Window, Viewer
except ImportError:
    Window, Viewer = None, None

INTERACTIVE = not hasattr(__main__, "__file__")


def _normalize_axes(axes: Iterable[Union[Axes, str]]) -> List[str]:
    """convert all axes to strings"""

    def _normalize(axes):
        if isinstance(axes, str):
            return axes
        elif isinstance(axes, Axes):
            return axes.value
        else:
            raise TypeError(f"passed axes must be str or Axes objects, not {type(axes)}")

    return list(_normalize(i) for i in axes)


def _max_intensity_table_maintain_dims(
    intensity_table: IntensityTable, dimensions: Set[Axes],
) -> IntensityTable:
    """
    Maximum project an IntensityTable over dimensions, retaining singletons in place of dimensions
    reduced by the max operation.

    For example, xarray.max(Axes.CH.value) would usually produce a 2-d (features, rounds) output.
    This function ensures that the output is instead (features, 1, rounds), by inserting singleton
    dimensions in place of any projected axes.

    Parameters
    ----------
    intensity_table : IntensityTable
        Input intensities
    dimensions : Set[Axes]
        Dimensions to project

    Returns
    -------
    IntensityTable :
        full-dimensionality (3-d) IntensityTable

    """
    str_dimensions = _normalize_axes(dimensions)
    initial_dimensions = OrderedDict(intensity_table.sizes)
    projected_intensities = intensity_table.max(str_dimensions)
    expanded_intensities = projected_intensities.expand_dims(str_dimensions)
    return expanded_intensities.transpose(*tuple(initial_dimensions.keys()))


def _mask_low_intensity_spots(
    intensity_table: IntensityTable,
    intensity_threshold: float
) -> np.ndarray:
    """
    returns a flattened boolean mask (c order flattening) that is True where
    (feature, round, channel) combinations whose intensities are above or equal to intensity
    threshold and where values are not nan
    """
    with warnings.catch_warnings():
        # we expect to have invalid values (np.nan) in IntensityTable, hide the warnings from users
        warnings.simplefilter("ignore", RuntimeWarning)
        # reshape the values into a 1-d vector of length product(features, rounds, channels)
        return np.ravel(intensity_table.values >= intensity_threshold)


def _spots_to_markers(intensity_table: IntensityTable) -> Tuple[np.ndarray, np.ndarray]:
    """
    convert each (r, c) combination for each spot into a 5-tuple (x, y, r, c, z) for plotting
    with napari.
    """
    # get sizes which will be used to create the coordinates that correspond to the spots
    n_rounds = intensity_table.sizes[Axes.ROUND.value]
    n_channels = intensity_table.sizes[Axes.CH.value]
    n_features = intensity_table.sizes[Features.AXIS]
    code_length = n_rounds * n_channels

    # create 5-d coordinates for plotting in (x, y, round. channel, z)
    n_markers = np.product(intensity_table.shape)
    coords = np.zeros((n_markers, 5), dtype=np.uint16)

    # create the coordinates.
    # X, Y, and Z are repeated once per (r, ch) pair (code_length).
    # the cartesian product of (r, c) are created, and tiled once per feature (n_features)
    # we ensure that (ch, round) cycle in c-order to match the order of the linearized
    # array, used below for masking.
    coords[:, 0] = np.repeat(intensity_table[Features.AXIS][Axes.X.value].values, code_length)
    coords[:, 1] = np.repeat(intensity_table[Features.AXIS][Axes.Y.value].values, code_length)
    coords[:, 2] = np.tile(np.tile(range(n_rounds), n_channels), n_features)
    coords[:, 3] = np.tile(np.repeat(range(n_channels), n_rounds), n_features)
    coords[:, 4] = np.repeat(intensity_table[Features.AXIS][Axes.ZPLANE.value].values, code_length)

    sizes = np.repeat(intensity_table.radius.values, code_length)
    xy = np.tile(sizes[:, np.newaxis], (1, 2))
    rc = np.zeros((sizes.shape[0], 2), dtype=int)
    z = sizes[:, np.newaxis]
    sizes = np.concatenate((xy, rc, z), axis=1)

    return coords, sizes


def display(
        stack: Optional[ImageStack] = None,
        spots: Optional[IntensityTable] = None,
        viewer: Optional[Viewer] = None,
        project_axes: Optional[Set[Axes]] = None,
        mask_intensities: float = 0.,
        radius_multiplier: int = 1,
        z_multiplier: float = 1
):
    """
    Displays an image stack and/or detected spots using Napari (https://github.com/napari/Napari).

    Parameters
    ----------
    stack : ImageStack
        ImageStack to display
    spots : IntensityTable
        IntensityTable containing spot information that was generated from the submitted stack.
    viewer : napari.Viewer
        Napari viewer to append the ImageStack and/or spots to. If None, creates a new viewer.
        Note: appending is only supported in interactive environments.
    project_axes : Optional[Set[Axes]]
        If provided, both the ImageStack and the Spots will be maximum projected along the
        selected axes. Useful for displaying spots across coded assays where spots may not
        appear in specific rounds or channels.
    mask_intensities : Float
        hide markers that correspond to intensities below this threshold value; note that any
        marker that is np.nan will ALSO be masked, allowing users to pre-mask the intensity table
        (see documentation on IntensityTable.where() for more details) (default 0, no masking)
    radius_multiplier : int
        Multiplies the radius of the displayed spots (default 1, no scaling)
    z_multiplier : int
        Multiplies the radius of the spots in z, to account for anisotropy.

    Examples
    --------

    1. Display a stack to evaluate a filtering result. Just pass any ImageStack!

    >>> starfish import display
    >>> display(stack)

    2. Display spots of a single-molecule FISH experiment: smFISH will produce IntensityTables where
    most values are np.NaN. These will be masked automatically, so passing an ImageStack +
    IntensityTable will display spots in the rounds and channels that they are detected

    >>> from starfish import display
    >>> display(stack, intensities)

    3. Diagnose spot calls within each round and channel of a coded experiment. A user might want to
    evaluate if spot calling is failing for a specific round/channel pair. To accomplish this,
    pass the intensity threshold used by the spot called to eliminate sub-threshold spots from
    channels. The user can additionally filter the IntensityTable to further mask additional spots
    (any np.NaN value will not be displayed)

    >>> from starfish import display
    >>> mask_intensities = 0.3  # this was the spot calling intensity threshold
    >>> display(stack, intensities, mask_intensities=mask_intensities)

    4. Evaluate spot calls across rounds and channels by visualizing spots on a max projection of
    The rounds and channels.

    >>> from starfish import display, Axes
    >>> display(stack, intensities, project_axes={Axes.CH, Axes.ROUND})

    5. Compare the image before (raw_stack) and after (filtered_stack) filtering by displaying
    two stacks in the same Viewer.

    >>> from starfish import display
    >>> viewer = display(raw_stack)
    >>> viewer = display(stack=filtered_stack, viewer=viewer)

    Notes
    -----
    - To use in ipython, use the `%gui qt5` magic.
    - Napari axes currently cannot be labeled. Until such a time that they can, this function will
      order them by Round, Channel, and Z.
    - Requires napari 0.0.6: install starfish using `pip install starfish[napari]` to install all
      necessary requirements
    """
    if stack is None and spots is None:
        raise TypeError("expected a stack and/or spots; got nothing")

    if Window is None or Viewer is None:
        warnings.warn("Requires napari 0.0.6. Run `pip install starfish[napari]` to install the "
                      "necessary requirements.")
        return

    from PyQt5.QtWidgets import QApplication

    # Instantiate the napari viewer
    if viewer is None:
        app = QApplication.instance() or QApplication([])

        # initialize the viewer
        viewer = Viewer()
        window = Window(viewer, show=False)
        viewer._window = window
        new_viewer = True
    elif isinstance(viewer, Viewer):
        new_viewer = False
    else:
        raise TypeError("viewer must be a napari Viewer or None")

    if stack is not None:
        if project_axes is not None:
            stack = stack.max_proj(*project_axes)

        # Switch axes to match napari expected order [x, y, round, channel, z]
        reordered_array: np.ndarray = stack.xarray.transpose(
            Axes.Y.value,
            Axes.X.value,
            Axes.ROUND.value,
            Axes.CH.value,
            Axes.ZPLANE.value
        ).values

        layer = viewer.add_image(np.squeeze(reordered_array), multichannel=False)
        layer.colormap = "gray"

    if spots is not None:
        if project_axes is not None:
            spots = _max_intensity_table_maintain_dims(spots, project_axes)

        # TODO ambrosejcarr guard rails:
        # 1. verify that x, y, z fit inside the imagestack (weird projections)
        # 2. warn if whole tiles are missing spot calls (possible miss-use)

        coords, sizes = _spots_to_markers(spots)

        # mask low-intensity values
        mask = _mask_low_intensity_spots(spots, mask_intensities)

        if not np.sum(mask):
            warnings.warn(f"No spots passed provided intensity threshold of {mask_intensities}")
            return viewer

        coords = coords[mask]
        sizes = sizes[mask]

        # adjust z-size
        sizes[:, 4] *= z_multiplier

        viewer.add_markers(coords=coords, face_color="red", edge_color="red", symbol="ring",
                           size=sizes * radius_multiplier, n_dimensional=True)

    if new_viewer:
        window.show()

        if not INTERACTIVE:
            app.exec()  # create blocking process to persist windows

    return viewer
